"""
Contains OrcaProc class for calculating ORCA related properties of conformers.
"""

import os
from pathlib import Path
import shutil
from typing import final, cast, override

from ..config.job_config import (
    NMRJobConfig,
    OptJobConfig,
    SPJobConfig,
    UVVisJobConfig,
    XTBOptJobConfig,
)
from .job import (
    JobContext,
)
from .results import (
    GsolvResult,
    NMRResult,
    OptResult,
    SPResult,
    MetaData,
    UVVisResult,
)
from ..utilities import Factory
from ..logging import setup_logger
from ..molecules import GeometryData
from ..params import (
    WARNLEN,
    USER_ASSETS_PATH,
    OrcaSolvMod,
    Prog,
)
from ..assets import FUNCTIONALS, SOLVENTS
from .qm_processor import QmProc

logger = setup_logger(__name__)


@final
class OrcaProc(QmProc):
    """
    Performs calculations with ORCA.
    """

    progname = Prog.ORCA

    # contains grid settings for ORCA 5.0+ (True) and older versions (False)
    # can be chosen by simple keyword (low/low+/high/high+)
    __gridsettings = {
        False: {
            "low": ["grid4", "nofinalgrid", "loosescf"],
            "low+": ["grid4", "nofinalgrid", "scfconv6"],
            "high": ["grid4", "nofinalgrid", "scfconv7"],
            "high+": ["grid5", "nofinalgrid", "scfconv7"],
            "nmr": ["grid5", "nofinalgrid", "scfconv7"],
        },
        True: {
            "low": ["DEFGRID1", "loosescf"],
            "low+": ["DEFGRID2", "scfconv6"],
            "high": ["DEFGRID2", "scfconv7"],
            "high+": ["DEFGRID2", "scfconv7"],
            "nmr": ["DEFGRID2", "scfconv7"],
        },
    }

    # Dict to map orca returncodes to error messages
    __returncode_to_err: dict[int, str] = {
        24: "input_error",
        25: "input_error",
        30: "input_error",
        52: "input_error",
        55: "unknown_error",
        125: "unknown_error",
    }

    def __prep(
        self,
        job: JobContext,
        config: SPJobConfig,
        jobtype: str,
        no_solv: bool = False,
        xyzfile: str | None = None,
    ) -> list[str]:
        """
        Prepares an OrderedDict to be fed into the parser in order to write an input file for jobtype 'jobtype'
        (e.g. sp).

        Can load up a template file from user assets folder.

        :param job: JobContext object containing the job information
        :type job: JobContext
        :param config: Configuration for the job
        :type config: SPJobConfig
        :param jobtype: jobtype to prepare the input for
        :type jobtype: str
        :param no_solv: if True, no solvent model is used
        :type no_solv: bool
        :param xyzfile: if not None, the geometry is read from this file instead of the job object
        :type xyzfile: str | None
        :returns: List of strings for the input file
        :rtype: list[str]

        NOTE: the xyzfile has to already exist, this function just bluntly writes the name into the input.
        """

        # check ORCA version (orca5 = True means at least ORCA version 5)
        # orca5 = int(config.paths.orcaversion[0]) > 4

        inp: list[str] = []

        template = "template" in config.model_fields and bool(config.template)
        if template:
            # NOTE: when using templates we're not going to check for double definitions!
            # if the template is messed up, orca will fail and the user should deal with that
            # load template file
            try:
                inp = (
                    (USER_ASSETS_PATH / f"{job.from_part}.orca.template")
                    .read_text()
                    .split("\n")
                )
            except FileNotFoundError:
                raise FileNotFoundError(
                    f"Could not find template file {job.from_part}.orca.template."
                )

        # prepare the main line of the orca input
        if template:
            main_line = next(inp.index(line) for line in inp if "{main}" in line)
            inp.pop(main_line)
        inp[:0] = self.__prep_main(jobtype, config, no_solv)

        # prepare all options that are supposed to be placed before the
        # geometry definition
        inp[1:1] = self.__prep_pregeom(jobtype, config, job.omp)

        # prepare the geometry
        if template:
            geom_line = next(inp.index(line) for line in inp if "{geom}" in line)
            inp.pop(geom_line)
            inp[geom_line:geom_line] = self.__prep_geom(
                job.conf,
                xyzfile,
                job.charge,
                job.unpaired,
            )
        else:
            inp.extend(
                self.__prep_geom(
                    job.conf,
                    xyzfile,
                    job.charge,
                    job.unpaired,
                )
            )

        inp.extend(self.__prep_postgeom(config, jobtype))

        # Finally append newlines where needed
        return [line + "\n" if not line.endswith("\n") else line for line in inp]

    def __prep_main(
        self, jobtype: str, config: SPJobConfig, no_solv: bool
    ) -> list[str]:
        orca5 = int(config.paths.orcaversion[0]) > 4

        # grab func, basis
        func = config.func
        basis = config.basis

        # All of this should be validated while setting up the config
        funcname = FUNCTIONALS[func][Prog.ORCA.value]
        functype = FUNCTIONALS[func]["type"]
        disp = FUNCTIONALS[func]["disp"]

        main = ["!", funcname]
        pregeom = []

        if "composite" not in functype:
            main.append(basis)

        # set  RI def2/J,   RIJCOSX def2/J

        # Set def2/J in case of def2 basis
        if "def2" in basis.lower():
            main.append("def2/J")
        # Otherwise use autoaux
        else:
            main.append("autoaux")

        # settings for double hybrids
        if "double" in functype:
            if "nmr" in jobtype:
                main.append("NOFROZENCORE")
                pregeom.extend(["%mp2", "RI true", "density relaxed", "end"])
            else:
                main.append("frozencore")
                pregeom.extend(["%mp2", "RI true", "end"])

            def2cbasis = ("def2-svp", "def2-tzvp", "def2-tzvpp", "def2-qzvpp")
            if basis.lower() in def2cbasis:
                main.append(f"{basis}/C")
            else:
                main.append("def2-tzvpp/C")

            if not orca5:
                main.extend(["GRIDX6", "NOFINALGRIDX"])
        # settings for hybrids
        elif "hybrid" in functype:
            if not orca5:
                main.extend(["GRIDX6", "NOFINALGRIDX"])

        # dummy type falls through every case, nothing is done in that case

        # use 'grid' setting from instructions to quickly configure the grid
        main.extend(self.__gridsettings[orca5][config.grid])

        # add dispersion
        # dispersion correction information
        # TODO: temporary solution (not very nice)
        mapping = {
            "d3bj": "d3bj",
            "d3(0)": "D3ZERO",
            "d4": "d4",
            "nl": "NL",
        }

        if disp != "composite" and disp != "included":
            main.append(mapping[disp])

        if disp == "nl" and not orca5:
            main.append("vdwgrid3")

        if "composite" not in functype:
            # try to apply gcp if basis set available
            # TODO - extend this
            gcp_keywords = {
                "minis": "MINIS",
                "sv": "SV",
                "6-31g(d)": "631GD",
                "def2-sv(p)": "SV(P)",
                "def2-svp": "SVP",
                "def2-tzvp": "TZ",
            }
            if basis.lower() in gcp_keywords.keys():
                main.append(f"GCP(DFT/{gcp_keywords[basis.lower()]})")

        # add job keyword for geometry optimizations
        # with ANCOPT
        if jobtype == "xtb_opt":
            main.extend(["ENGRAD", "tightSCF"])
        # for standard geometry optimization
        elif jobtype == "opt":
            main.extend(["OPT", "tightSCF"])

        # set keywords for the selected solvent model
        if not no_solv:
            assert config.solvent
            assert config.sm
            sm = config.sm
            solv_key = f"{SOLVENTS[config.solvent][sm]}"

            orca6 = int(config.paths.orcaversion[0]) > 5
            if orca6 or sm != OrcaSolvMod.SMD:
                main.append(f"{sm.upper()}({solv_key})")
            else:
                pregeom.extend(["%cpcm", "smd true", f'smdsolvent "{solv_key}"', "end"])

        # additional print settings
        if "opt" in jobtype:
            main.append("miniprint")
        else:
            main.append("printgap")

        return [" ".join(main)] + pregeom

    def __prep_pregeom(
        self,
        jobtype: str,
        config: SPJobConfig,
        nprocs: int,
    ) -> list[str]:
        orca5 = int(config.paths.orcaversion[0]) > 4
        pregeom = []

        # Check ORCA version (important for grid keywords)
        if orca5 and nprocs > 1:
            pregeom.extend(["%pal", f"nprocs {nprocs}", "end"])

        # TODO: maybe limit TRAH macro steps (or when TRAH activates) to avoid
        # single jobs clogging everything up

        if jobtype == "uvvis":
            assert isinstance(config, UVVisJobConfig)
            pregeom.extend(["%tddft", f"nroots {config.nroots}", "end"])

        # Additional print settings
        if "opt" not in jobtype:
            pregeom.extend(["%output", "printlevel normal", "end"])

        if jobtype == "opt":
            assert isinstance(config, OptJobConfig)
            if config.macrocycles:
                # Set max number of optimization cycles for ORCA driven optimization
                pregeom.extend(["%geom", f"maxiter {config.optcycles}"])
            else:
                pregeom.extend(["%geom"])

            # Set optlevel
            # NOTE: this is only approximatively equivalent,
            # convergence criteria differ too much between ANCOPT and ORCA
            mapping = {
                "crude": "loose",
                "sloppy": "loose",
                "loose": "loose",
                "lax": "loose",
                "normal": "loose",
                "tight": "normal",
                "vtight": "tight",
                "extreme": "tight",
            }

            # Try to apply literally first
            if config.optlevel in ["loose", "normal", "tight"]:
                pregeom.extend([f"convergence {config.optlevel}", "end"])
            # Otherwise map to roughly corresponding orca optlevel
            else:
                pregeom.extend([f"convergence {mapping[config.optlevel]}", "end"])

            # Insert constraints (if provided)
            # FIXME: without better parser this will not work
            """
            if prepinfo["opt"]["constraints"] is not None:
                parser = OrcaParser()
                constraints = parser.read_input(prepinfo["opt"]["constraints"])
                indict["geom"].update(constraints["geom"])
            """

        return pregeom

    def __prep_geom(
        self,
        conf: GeometryData,
        xyzfile: str | None,
        charge: int,
        unpaired: int,
    ) -> list[str]:
        lines: list[str] = []
        # unpaired, charge, and coordinates
        # by default coordinates are written directly into input file
        if xyzfile is None:
            lines.extend([f"* xyz {charge} {unpaired + 1}"] + conf.toorca())
            lines.append("*")
        else:
            lines.extend([f"* xyzfile {charge} {unpaired + 1} {xyzfile}"])

        return lines

    def __prep_postgeom(
        self,
        config: SPJobConfig,
        jobtype: str,
    ) -> list[str]:
        lines = []
        # Set NMR parameters
        if "nmr" in jobtype:
            assert isinstance(config, NMRJobConfig)
            # Determine the settings that need to be put into the input file for the NMR calculation
            todo = config.active_nuclei.split(",")

            todo2 = []
            lines.append("%eprnmr")
            if config.shieldings:
                todo2.append("shift")
                lines.extend(
                    [
                        "origin giao",
                        "giao_2el giao_2el_same_as_scf",
                        "giao_1el giao_1el_analytic",
                    ]
                )
            if config.couplings:
                if config.fc_only:
                    todo2.append("ssfc")
                else:
                    todo2.append("ssall")
                lines.append(f"SpinSpinRThresh {config.ss_cutoff:.4f}")

            lines.extend(
                [
                    f"Nuclei = all {element.capitalize()} "
                    + "{ "
                    + ",".join(todo2)
                    + " }"
                    for element in todo
                ]
            )
            lines.append("end")

        return lines

    @staticmethod
    def __check_output(lines: list[str]) -> str:
        """
        Checks the lines from the output file for errors and returns them.

        :param lines: list of lines from the output file.
        :type lines: list[str]
        :returns: error message if an error was found, empty string otherwise
        :rtype: str
        """
        # Dict mapping specific messages from the output to error messages
        # TODO: this should be extended later
        out_to_err = {
            "SCF NOT CONVERGED": "scf_not_converged",
        }
        for line in lines:
            if any(key in line for key in out_to_err.keys()):
                # Returns the first error found
                key = next(filter(lambda x: x in line, out_to_err.keys()))
                return out_to_err[key]
        return ""

    @staticmethod
    def __copy_mo(jobdir: str | Path, filename: str, guess_file: str | Path) -> None:
        """
        Copy MO file if possible (should be ORCA .gbw file).

        :param jobdir: path to the job directory
        :type jobdir: str | Path
        :param filename: name of the input file
        :type filename: str
        :param guess_file: path to the .gbw file to copy
        :type guess_file: str | Path
        :returns: None
        :rtype: None
        """
        if os.path.isfile(guess_file) and ".gbw" in os.path.split(guess_file)[1]:
            if os.path.join(jobdir, f"{filename}.gbw") != guess_file:
                logger.debug(
                    f"{f'worker{os.getpid()}:':{WARNLEN}}Copying .gbw file from {guess_file}."
                )
                shutil.copy(guess_file, os.path.join(jobdir, f"{filename}.gbw"))

    @final
    def _sp(
        self,
        job: JobContext,
        config: SPJobConfig,
        jobdir: str | Path | None = None,
        filename: str = "sp",
        no_solv: bool = False,
        prep: bool = True,
    ) -> tuple[SPResult, MetaData]:
        """
        ORCA single-point calculation.
        Unwrapped function to call from other methods.

        :param job: JobContext object containing the job information, metadata is stored in job.meta
        :type job: JobContext
        :param config: Configuration for the job
        :type config: SPJobConfig
        :param jobdir: path to the job directory
        :type jobdir: str | Path | None
        :param filename: name of the input file
        :type filename: str
        :param no_solv: if True, no solvent model is used
        :type no_solv: bool
        :param prep: if True, a new input file is generated (you only really want to make use of this for NMR)
        :type prep: bool
        :returns: Tuple of (SPResult, MetaData)
        :rtype: tuple[SPResult, MetaData]
        """
        if jobdir is None:
            jobdir = self._setup(job, "sp")

        # set results
        result = SPResult()
        meta = MetaData(job.conf.name)

        # set in/out path
        inputpath = os.path.join(jobdir, f"{filename}.inp")
        outputpath = os.path.join(jobdir, f"{filename}.out")

        # prepare input
        if prep:
            inp = self.__prep(job, config, "sp", no_solv=config.gas_phase or no_solv)

            # write input into file "{filename}.inp" in a subdir created for the
            # conformer
            with open(inputpath, "w") as f:
                f.writelines(inp)

        # check, if there is an existing .gbw file and copy it if option
        # 'copy_mo' is true
        if (
            config.copy_mo
            and job.mo_guess is not None
            and not isinstance(job.mo_guess, tuple)
        ):
            self.__copy_mo(jobdir, filename, job.mo_guess)

        # call orca
        call = [config.paths.orca, f"{filename}.inp"]
        returncode, _ = self._make_call(call, outputpath, jobdir)
        # NOTE: using orca returncodes it is not possible to determine wether the calculation converged

        meta.success = returncode == 0

        # read output
        with open(outputpath) as out:
            lines = out.readlines()

        # Check for errors in the output file in case returncode is 0
        if meta.success:
            meta.error = self.__check_output(lines)
            if meta.error != "":
                return result, meta
        else:
            meta.error = self.__returncode_to_err.get(returncode, "unknown_error")
            return result, meta

        # Get final energy
        try:
            result.energy = next(
                (
                    float(line.split()[4])
                    for line in lines
                    if "FINAL SINGLE POINT ENERGY" in line
                ),
            )
        except StopIteration:
            meta.success = False
            meta.error = "Could not parse final energy"

        if config.copy_mo:
            # store the path to the current .gbw file for this conformer if possible
            if os.path.isfile(os.path.join(jobdir, f"{filename}.gbw")):
                result.mo_path = os.path.join(jobdir, f"{filename}.gbw")

        return result, meta

    @override
    def sp(self, *args, **kwargs):
        """
        Perform single-point calculation.

        :param args: Arguments.
        :param kwargs: Keyword arguments.
        :return: Tuple of (SP result, metadata).
        """
        return self._sp(*args, **kwargs)

    @override
    def gsolv(
        self, job: JobContext, config: SPJobConfig
    ) -> tuple[GsolvResult, MetaData]:
        """
        Calculates the solvation free enthalpy of a conformer using ORCA.

        :param job: JobContext object containing the job information, metadata is stored in job.meta
        :param config: SP configuration
        :return: Tuple of (GsolvResult, MetaData)
        """
        # Check required settings

        # what is returned in the end
        result = GsolvResult()
        meta = MetaData(job.conf.name)

        jobdir = self._setup(job, "gsolv")

        # calculate gas phase
        spres, spmeta = self._sp(
            job, config, jobdir=jobdir, filename="sp_gas", no_solv=True
        )

        if spmeta.success:
            result.energy_gas = spres.energy
        else:
            meta.success = False
            meta.error = spmeta.error
            return result, meta

        # calculate in solution
        spres, spmeta = self._sp(job, config, jobdir=jobdir, filename="sp_solv")

        if spmeta.success:
            result.energy_solv = spres.energy
        else:
            meta.success = False
            meta.error = spmeta.error
            return result, meta

        if config.copy_mo:
            # store the path to the current .gbw file for this conformer if
            # possible
            assert isinstance(spres.mo_path, str | Path)
            if os.path.isfile(spres.mo_path):
                result.mo_path = spres.mo_path

        # calculate solvation free enthalpy
        result.gsolv = result.energy_solv - result.energy_gas
        meta.success = True

        return result, meta

    @override
    def opt(
        self,
        job: JobContext,
        config: OptJobConfig,
    ) -> tuple[OptResult, MetaData]:
        """
        Geometry optimization using ORCA optimizer.
        Note that solvation in handled here always implicitly.

        :param job: JobContext object containing the job information, metadata is stored in job.meta
        :param config: Optimization configuration
        :return: Tuple of (OptResult, MetaData)
        """
        # prepare result
        result = OptResult()
        meta = MetaData(job.conf.name)

        jobdir = self._setup(job, "opt")
        filename = "opt"

        # set orca input/output paths
        inputpath = os.path.join(jobdir, f"{filename}.inp")
        outputpath = os.path.join(jobdir, f"{filename}.out")

        # prepare input dict
        # TODO: add constraints
        inp = self.__prep(job, config, "opt", no_solv=config.gas_phase)

        # write input into file "{filename}.inp" in a subdir created for the
        # conformer
        with open(inputpath, "w") as f:
            f.writelines(inp)

        # Get gbw files for initial guess
        if (
            config.copy_mo
            and job.mo_guess is not None
            and not isinstance(job.mo_guess, tuple)
        ):
            self.__copy_mo(jobdir, filename, job.mo_guess)

        # call orca
        call = [config.paths.orca, f"{filename}.inp"]
        returncode, _ = self._make_call(call, outputpath, jobdir)
        # NOTE: using orca returncodes it is not possible to determine wether the calculation converged

        meta.success = returncode == 0

        # read output
        with open(outputpath) as out:
            lines = out.readlines()

        # Check for errors in the output file in case returncode is 0
        if meta.success:
            meta.error = self.__check_output(lines)
            if meta.error != "":
                return result, meta
        else:
            meta.error = self.__returncode_to_err.get(returncode, "unknown_error")
            return result, meta

        # Check convergence
        if (
            next((True for x in lines if "OPTIMIZATION HAS CONVERGED" in x), None)
            is True
        ):
            result.converged = True
        else:
            result.converged = False

        # Get the number of cycles
        for line in lines:
            if "GEOMETRY OPTIMIZATION CYCLE" in line:
                result.cycles = int(line.split()[4])

        # Get energies for each cycle
        result.ecyc = [
            float(line.split("....")[-1].split()[0])
            for line in filter(lambda x: "Current Energy" in x, lines)
        ]

        # Get all gradient norms for evaluation
        result.gncyc = [
            float(line.split("....")[-1].split()[0])
            for line in filter(lambda x: "Current gradient norm" in x, lines)
        ]

        # Get the last gradient norm and energy
        result.energy = result.ecyc[-1]
        result.grad_norm = result.gncyc[-1]
        meta.success = True

        # Read out optimized geometry and update conformer geometry with this
        job.conf.fromxyz(os.path.join(jobdir, f"{filename}.xyz"))
        result.geom = job.conf.xyz

        if config.copy_mo:
            # store the path to the current .gbw file for this conformer if
            # possible
            if os.path.isfile(os.path.join(jobdir, f"{filename}.gbw")):
                result.mo_path = os.path.join(jobdir, f"{filename}.gbw")

        return result, meta

    @override
    def xtb_opt(
        self,
        job: JobContext,
        config: XTBOptJobConfig,
    ) -> tuple[OptResult, MetaData]:
        """
        Geometry optimization using ANCOPT and ORCA gradients.
        Note that solvation is handled here always implicitly.

        :param job: JobContext object containing the job information, metadata is stored in job.meta
        :param config: XTB optimization configuration
        :return: Tuple of (OptResult, MetaData)
        """
        # NOTE: some "intuitivity problems":
        # the geometry of the conformer is written into a coord file and also into an xyz-file to be used by orca
        # xtb then outputs a file with the optimized geometry as 'xtbopt.coord', which is then read into the conformer
        # to update it's geometry

        # prepare result
        # 'ecyc' contains the energies for all cycles, 'cycles' stores the number of required cycles
        # 'gncyc' contains the gradient norms for all cycles
        # 'energy' contains the final energy of the optimization (converged or unconverged)
        # 'geom' stores the optimized geometry in GeometryData.xyz format
        result = OptResult()
        meta = MetaData(job.conf.name)
        xcontrolname = "xtb_opt-xcontrol-inp"

        files = [
            "xtbrestart",
            "xtbtopo.mol",
            xcontrolname,
            "wbo",
            "charges",
            "gfnff_topo",
        ]

        jobdir = self._setup(job, "xtb_opt")
        filename = "xtb_opt"

        # remove potentially preexisting files to avoid confusion
        for file in files:
            if os.path.isfile(os.path.join(jobdir, file)):
                os.remove(os.path.join(jobdir, file))

        # write conformer geometry to coord file
        with open(os.path.join(jobdir, f"{filename}.coord"), "w", newline=None) as f:
            f.writelines(job.conf.tocoord())

        # write xyz-file for orca
        with open(os.path.join(jobdir, f"{filename}.xyz"), "w", newline=None) as f:
            f.writelines(job.conf.toxyz())

        # set orca input path
        inputpath = os.path.join(jobdir, f"{filename}.inp")

        # prepare input dict
        inp = self.__prep(job, config, "xtb_opt", xyzfile=f"{filename}.xyz")

        # write orca input into file "xtb_opt.inp" in a subdir created for the
        # conformer
        with open(inputpath, "w") as f:
            f.writelines(inp)

        # append some additional lines to the coord file for ancopt
        with open(
            os.path.join(jobdir, f"{filename}.coord"), "a", newline=None
        ) as newcoord:
            newcoord.writelines(
                [
                    "$external\n",
                    f"   orca input file= {filename}.inp\n",
                    f"   orca bin= {config.paths.orca}\n",
                    "$end\n",
                ]
            )

        # prepare configuration file for ancopt (xcontrol file)
        with open(os.path.join(jobdir, xcontrolname), "w", newline=None) as out:
            out.write("$opt \n")
            if config.macrocycles:
                out.write(f"maxcycle={config.optcycles} \n")
                out.write(f"microcycle={config.optcycles} \n")

            out.writelines(
                [
                    "average conv=true \n",
                    f"hlow={config.hlow} \n",
                    "s6=30.00 \n",
                    "engine=lbfgs\n",
                    "$external\n",
                    f"   orca input file= {filename}.inp\n",
                    f"   orca bin= {config.paths.orca} \n",
                ]
            )

            # Import constraints
            if config.constraints:
                out.writelines(config.constraints)

            out.write("$end \n")

        # check, if there is an existing .gbw file and copy it if option
        # 'copy_mo' is true
        if (
            config.copy_mo
            and job.mo_guess is not None
            and not isinstance(job.mo_guess, tuple)
        ):
            self.__copy_mo(jobdir, filename, job.mo_guess)

        # prepare xtb call
        call = [
            config.paths.xtb,
            f"{filename}.coord",  # name of the coord file generated above
            "--opt",
            config.optlevel,
            "--orca",
            "-I",
            xcontrolname,
        ]

        # set path to the ancopt output file
        outputpath = os.path.join(jobdir, f"{filename}.out")

        # call xtb
        returncode, _ = self._make_call(call, outputpath, jobdir)

        # check if optimization finished without errors
        # NOTE: right now, not converging scfs are not handled because returncodes need to be implemented first
        if returncode != 0:
            meta.success = False
            meta.error = "unknown_error"
            # TODO: the xtb returncodes should be handled
            return result, meta

        # read output
        with open(outputpath) as f:
            lines = f.readlines()

        result.ecyc = []
        result.cycles = 0

        # Substrings indicating error in xtb
        error_ind = [
            "external code error",
            "|grad| > 500, something is totally wrong!",
            "abnormal termination of xtb",
        ]

        # Check if xtb terminated normally (if there are any error indicators
        # in the output)
        meta.success = (
            False
            if next((x for x in lines if any(y in x for y in error_ind)), None)
            is not None
            else True
        )
        if not meta.success:
            meta.error = "unknown_error"
            return result, meta

        # check convergence
        if (
            next((True for x in lines if "GEOMETRY OPTIMIZATION CONVERGED" in x), None)
            is True
        ):
            result.converged = True
        elif (
            next((True for x in lines if "FAILED TO CONVERGE GEOMETRY" in x), None)
            is True
        ):
            result.converged = False

        # Get the number of cycles
        # tmp is one of the values from the dict defined below
        tmp_val = {
            True: ("GEOMETRY OPTIMIZATION CONVERGED", 5),
            False: ("FAILED TO CONVERGE GEOMETRY", 7),
        }[result.converged]

        result.cycles = int(
            next(x for x in lines if tmp_val[0] in x).split()[tmp_val[1]]
        )

        # Get energies for each cycle
        result.ecyc.extend(
            float(line.split("->")[-1])
            for line in filter(lambda x: "av. E: " in x, lines)
        )

        # Get all gradient norms for evaluation
        result.gncyc = [
            float(line.split()[3])
            for line in filter(lambda x: " gradient norm " in x, lines)
        ]

        # Get the last gradient norm
        result.grad_norm = result.gncyc[-1]

        # store the final energy of the optimization in 'energy'
        result.energy = result.ecyc[-1]
        meta.success = True

        if config.copy_mo:
            # store the path to the current .gbw file for this conformer
            result.mo_path = os.path.join(jobdir, f"{filename}.gbw")

        # read out optimized geometry and update conformer geometry with this
        job.conf.fromcoord(os.path.join(jobdir, "xtbopt.coord"))
        result.geom = job.conf.xyz

        return result, meta

    @override
    def nmr(
        self,
        job: JobContext,
        config: NMRJobConfig,
    ) -> tuple[NMRResult, MetaData]:
        """
        Calculate the NMR shieldings and/or couplings for a conformer using ORCA. ORCA gives only the active cores in the output
        so there is not need for more thinking here.
        Formatting:
            'shielding' contains a list of tuples (atom_index, shielding), with atom_index being the index of the atom
            in the internal coordinates of the GeometryData.
            'couplings' contains a list of tuples ((atom_index1, atom_index2), coupling), with the indices of the atoms
            in the internal coordinates of the GeometryData. A set is used to represent an atom pair and then converted
            to tuple to be serializable.

        :param job: JobContext object containing the job information, metadata is stored in job.meta
        :param config: NMR configuration
        :return: Tuple of (NMRResult, MetaData)
        """
        # Set results
        result = NMRResult()
        meta = MetaData(job.conf.name)

        jobdir = self._setup(job, "nmr")
        filename = "nmr"

        # Set in/out path
        inputpath = os.path.join(jobdir, f"{filename}.inp")
        outputpath = os.path.join(jobdir, f"{filename}.out")

        # Prepare an input file for _sp
        inp = self.__prep(job, config, "nmr", no_solv=config.gas_phase)
        with open(inputpath, "w") as f:
            f.writelines(inp)

        # Run _sp using the input file generated above
        _, spmeta = self._sp(job, config, jobdir=jobdir, filename=filename, prep=False)

        if not spmeta.success:
            meta.success = False
            meta.error = spmeta.error
            return result, meta

        # Grab shieldings and energy from the output
        with open(outputpath) as f:
            lines = f.readlines()

        # Get final energy
        # result.energy = spres.energy

        # For shieldings watch out for the line "CHEMICAL SHIELDING SUMMARY
        # (ppm)"
        start = (
            lines.index(next(x for x in lines if "CHEMICAL SHIELDING SUMMARY" in x)) + 6
        )

        result.shieldings = []

        for line in lines[start:]:
            # Try to extract values from the lines, if that fails
            # (probably IndexError) stop
            try:
                result.shieldings.append((int(line.split()[0]), float(line.split()[2])))
            except IndexError:
                break

        # Sort shieldings by atom index
        result.shieldings.sort(key=lambda x: x[0])

        with open(outputpath) as f:
            lines = f.readlines()

        start = lines.index(next(x for x in lines if "SPIN-SPIN COUPLING" in x)) + 6

        lines = lines[start:]

        end = lines.index(next(x for x in lines if "SUMMARY OF ISOTROPIC" in x)) - 3

        lines = lines[:end]

        couplings: list[tuple[frozenset[int], float]] = []

        # This goes through all the pairs, even though ORCA gives also non-unique pairs, since it makes life
        # easier (basically every element of a symmetric square matrix)
        # The iterator created here unpacks a tuple consisting of the multiples of three (indices of every
        # third line) and every third line in 'lines'
        for i, line in enumerate(lines):
            if "NUCLEUS" in line:
                # pair needs to be a frozenset because normal sets are not hashable and can therefore not be part
                # of a normal set
                pair = frozenset((int(line.split()[4]), int(line.split()[9])))
                for line2 in lines[i:]:
                    if "(Total)" in line2:
                        coupling = float(line2.split()[5])
                        couplings.append((pair, coupling))
                        break

        # Convert to set and back to get rid of duplicates
        # ('vectorizing the symmetric matrix')
        couplings = list(set(couplings))

        # Convert all the frozensets to a tuple to be serializable
        for i in range(len(couplings)):
            pair_tuple = cast(tuple[int, int], tuple(couplings[i][0]))
            coupling = couplings[i][1]
            result.couplings.append(
                (
                    pair_tuple,
                    couplings[i][1],
                )
            )

        # Sort couplings by pairs
        result.couplings.sort(key=lambda x: x[0])

        meta.success = True

        return result, meta

    @override
    def uvvis(
        self,
        job: JobContext,
        config: UVVisJobConfig,
    ) -> tuple[UVVisResult, MetaData]:
        """
        Run a single-point to calculate the oscillator strengths and excitation wavelengths.

        :param job: JobContext object containing the job information, metadata is stored in job.meta
        :param config: UV-Vis configuration
        :return: Tuple of (UVVisResult, MetaData)
        """
        # Set results
        result = UVVisResult()
        meta = MetaData(job.conf.name)

        jobdir = self._setup(job, "uvvis")
        filename = "uvvis"

        # Set in/out path
        inputpath = os.path.join(jobdir, f"{filename}.inp")
        outputpath = os.path.join(jobdir, f"{filename}.out")

        # Prepare an input file for _sp
        inp = self.__prep(job, config, "uvvis", no_solv=config.gas_phase)
        with open(inputpath, "w") as f:
            f.writelines(inp)

        # Run _sp using the input file generated above
        _, spmeta = self._sp(
            job, config, jobdir=jobdir, filename=f"{filename}", prep=False
        )

        if not spmeta.success:
            meta.success = False
            meta.error = spmeta.error
            return result, meta

        # Grab excitations and energy from the output
        with open(outputpath) as f:
            lines = f.readlines()

        # Get final energy
        # result.energy = spres.energy

        # Find the line index of the actual excitations information
        start = lines.index(
            next(filter(lambda line: "ABSORPTION SPECTRUM" in line, lines))
        )

        # Get all the lines where the excitations are listed (+5 because of spacing in output file)
        uvvis_table = lines[start + 5 : start + 5 + config.nroots]

        # Extract excitation wavelengths and oscillator strengths
        result.excitations = []
        for row in uvvis_table:
            spl = row.split()
            result.excitations.append(
                {"wavelength": float(spl[-6]), "osc_str": float(spl[-5])}
            )

        meta.success = True

        return result, meta

    @override
    def rot(self, *args, **kwargs):
        raise NotImplementedError


Factory.register_builder(Prog.ORCA, OrcaProc)



import enum
from pydantic import BaseModel, ConfigDict, field_serializer


class GenericConfig(BaseModel):
    """
    Generic configuration base class using Pydantic for validation and serialization.
    """

    model_config = ConfigDict(
        # use_enum_values=True,  # this might lead to problems somewhere
        str_to_lower=True,
        str_strip_whitespace=True,
        validate_default=True,
        use_attribute_docstrings=True,  # , validate_assignment=True <- don't use this for now
    )

    @field_serializer("*", when_used="json-unless-none")
    def serialize_enums(self, value):
        """
        Serialize enum values to their string representation for JSON output.

        :param value: The value to serialize.
        :return: The serialized value, with enums converted to their values.
        """
        if isinstance(value, enum.Enum):
            return value.value
        return value



"""Tests for PartsConfig"""

import pytest
from censo.config.parts_config import PartsConfig
from censo.config.parts import (
    GeneralConfig,
    PrescreeningConfig,
    ScreeningConfig,
    OptimizationConfig,
    RefinementConfig,
    NMRConfig,
    UVVisConfig,
)
from censo.params import OrcaSolvMod, TmSolvMod


def test_parts_config_default_initialization():
    """Test that PartsConfig initializes with default values"""
    config = PartsConfig()

    # Test that all parts are initialized with their default configurations
    assert isinstance(config.general, GeneralConfig)
    assert isinstance(config.prescreening, PrescreeningConfig)
    assert isinstance(config.screening, ScreeningConfig)
    assert isinstance(config.optimization, OptimizationConfig)
    assert isinstance(config.refinement, RefinementConfig)
    assert isinstance(config.nmr, NMRConfig)
    assert isinstance(config.uvvis, UVVisConfig)


def test_parts_config_str_representation():
    """Test the string representation of PartsConfig"""
    config = PartsConfig()
    str_repr = str(config)

    # Verify that the string representation contains all parts
    assert str_repr.count("\n") > 0  # Should have multiple lines
    for part_name in [
        "general",
        "prescreening",
        "screening",
        "optimization",
        "refinement",
        "nmr",
        "uvvis",
    ]:
        assert part_name in str_repr.lower()


@pytest.mark.parametrize(
    "solvent,sm_model,should_pass",
    [
        ("water", OrcaSolvMod.CPCM, True),  # Valid combination
        ("chloroform", OrcaSolvMod.SMD, True),  # Valid combination
        ("invalid_solvent", OrcaSolvMod.CPCM, False),  # Invalid solvent
    ],
)
def test_solvent_model_validation(solvent, sm_model, should_pass):
    """Test solvent model validation with different combinations"""
    config = PartsConfig()
    config.general.solvent = solvent

    # Set the same solvent model for all parts that use it
    config.screening.sm = sm_model
    config.optimization.sm = sm_model
    config.refinement.sm = sm_model
    config.nmr.sm = sm_model
    config.uvvis.sm = sm_model

    if should_pass:
        # Should not raise any validation errors
        config.model_validate(config, context={"check_all": True, "check_paths": False})
    else:
        # Should raise ValueError for invalid combinations
        with pytest.raises(ValueError, match="not available with"):
            config.model_validate(config, context={"check_all": True})


def test_custom_config_values():
    """Test PartsConfig with custom values"""
    custom_config = PartsConfig(
        general=GeneralConfig(solvent="water"),
        screening=ScreeningConfig(sm=OrcaSolvMod.CPCM),
        optimization=OptimizationConfig(sm=OrcaSolvMod.CPCM),
    )

    assert custom_config.general.solvent == "water"
    assert custom_config.screening.sm == OrcaSolvMod.CPCM
    assert custom_config.optimization.sm == OrcaSolvMod.CPCM


def test_invalid_solvent_combination():
    """Test that invalid solvent/model combinations raise appropriate errors"""
    config = PartsConfig()
    config.general.solvent = "dmf"
    config.screening.sm = OrcaSolvMod.CPCM
    config.optimization.sm = TmSolvMod.DCOSMORS  # Different solvent model

    # Should raise ValueError due to incompatible solvent/model combination
    with pytest.raises(ValueError):
        config.model_validate(config, context={"check_all": True})


def test_paths_model_validation():
    """Test that missing paths raise proper validation errors when check_paths=True"""
    config = PartsConfig()
    # Do not set required paths; assume default config is missing some required paths
    with pytest.raises(ValueError, match="path is not set in the configuration"):
        config.model_validate(config, context={"check_all": True, "check_paths": True})


def test_parts_validation():
    config = PartsConfig()
    config.screening.func = "invalid"
    with pytest.raises(ValueError):
        config.model_validate(config, context={"check": "screening"})

    config = PartsConfig()
    config.screening.gsolv_included = True
    config = config.model_validate(config, context={"check": "screening"})
    assert not config.screening.gsolv_included



from pathlib import Path
from typing import Union

from ..molecules import GeometryData


class JobContext:
    """
    Represents a job for parallel execution.
    """

    def __init__(self, conf: GeometryData, charge: int, unpaired: int, omp: int):
        """
        Initialize parallel job.

        :param conf: Geometry data.
        :param charge: Charge.
        :param unpaired: Unpaired electrons.
        :param omp: Number of cores.
        """
        # conformer for the job
        self.conf: GeometryData = conf

        # number of cores to use
        self.omp: int = omp

        # stores path to an mo file which is supposed to be used as a guess
        # In case of open shell tm calculation this can be a tuple of files
        self.mo_guess: Union[
            Path, str, tuple[Union[str, Path], Union[str, Path]], None
        ] = None

        self.from_part: str = ""

        self.charge: int = charge
        self.unpaired: int = unpaired

        # stores all flags for the jobtypes
        self.flags: dict[str, str] = {}



from typing import Literal, Any, Self
import ast
from pydantic import model_validator, Field, field_validator

from .base import BasePartConfig
from ...assets import FUNCTIONALS
from ...params import QmProg


class RotConfig(BasePartConfig):
    """Config class for optical rotation calculations (gas-phase only)."""

    prog: Literal[QmProg.TM] = QmProg.TM
    """Program that should be used for calculations."""

    func: str = "pbe-d4"
    """Functional that should be used for calculations."""

    basis: str = "def2-SVPD"
    """Basis set that should be used for calculations."""

    freq: list[float] = Field(default_factory=lambda: [589.0, 633.0])
    """List of frequencies for which optical rotation will be calculated."""

    template: bool = False
    """Whether to use template files."""

    @field_validator("freq")
    @classmethod
    def freq_must_be_valid(cls, v: list[float]):
        """
        Validate that frequencies are positive and non-empty.

        :param v: The list of frequencies.
        :return: The validated list.
        """
        if len(v) == 0:
            raise ValueError("No frequencies provided.")
        if any(x <= 0 for x in v):
            raise ValueError("Negative or 0 frequencies not allowed.")

        return v

    @field_validator("freq", mode="before")
    @classmethod
    def freq_cast(cls, v: Any):
        """
        Cast string input to list of floats.

        :param v: The input value.
        :return: The casted list.
        """
        if isinstance(v, str):
            parsed = ast.literal_eval(v)
            if not isinstance(parsed, list) or not all(
                isinstance(x, float) or isinstance(x, int) for x in parsed
            ):
                raise ValueError(
                    f"No valid list for optical rotation frequencies provided ({v}). Must have correct list syntax and contain only numeric values."
                )
            return parsed
        return v

    @model_validator(mode="after")
    def func_must_be_known_in_prog(self) -> Self:
        """
        Validate that the functional is known for the chosen program.

        :return: The validated instance.
        """
        prog: str = self.prog
        try:
            assert FUNCTIONALS[self.func][prog] is not None
            assert FUNCTIONALS[self.func]["disp"] is not None
            assert FUNCTIONALS[self.func]["type"] is not None
        except (KeyError, AssertionError):
            raise ValueError(
                f"Functional {self.func} not (fully) defined for prog {prog}."
            )
        return self



"""Tests for BasePartConfig"""

import enum
import warnings

from censo.config.parts.base import BasePartConfig
from censo.params import PLENGTH, QmProg


def test_base_part_config_str():
    """Test the string representation of BasePartConfig"""

    class TestEnum(enum.Enum):
        """Test enum"""

        A = "a"
        B = "b"

    class TestBaseConfig(BasePartConfig):
        """Test implementation of BasePartConfig"""

        test_value: str = "test"
        test_number: int = 42
        test_enum: TestEnum = TestEnum.A

    config = TestBaseConfig()
    result = str(config)

    # Check header formatting
    assert "TestBase" in result

    # Check value formatting
    assert "test_value" in result
    assert "test" in result
    assert "test_number" in result
    assert "42" in result
    assert "TestEnum" not in result

    # Check alignment
    lines = result.split("\n")
    for line in lines[1:]:  # Skip header
        if ":" in line:
            name = line.split(":")[0]
            assert len(name) < PLENGTH // 2


def test_tm_gcp_d4_bug_check():
    """Test the tm_gcp_d4_bug_check validator"""

    class TestConfigWithFields(BasePartConfig):
        """Test config with func, basis, prog fields"""

        func: str = "pbe-d4"
        basis: str = "def2-sv(p)"
        prog: QmProg = QmProg.TM

    # Test case that triggers the warning and correction
    with warnings.catch_warnings(record=True) as w:
        warnings.simplefilter("always")
        config = TestConfigWithFields(basis="def2-sv(p)", func="pbe-d4", prog=QmProg.TM)
        assert config.func == "pbe-d3"  # Should be corrected
        assert len(w) == 1
        assert "Small basis set detected" in str(w[0].message)

    # Test case that does not trigger (different basis)
    config2 = TestConfigWithFields(basis="def2-qzvp", func="pbe-d4", prog=QmProg.TM)
    assert config2.func == "pbe-d4"  # Should not be changed

    # Test case that does not trigger (different disp)
    config3 = TestConfigWithFields(basis="def2-sv(p)", func="pbe-d3", prog=QmProg.TM)
    assert config3.func == "pbe-d3"  # Should not be changed

    # Test case that does not trigger (different prog)
    config4 = TestConfigWithFields(basis="def2-sv(p)", func="pbe-d4", prog=QmProg.ORCA)
    assert config4.func == "pbe-d4"  # Should not be changed



"""
defininition of internal defaults, checking of logic for parameter combinations,
cml parsing
"""

from ..params import START_DESCR
import argparse

from ..params import OMPMIN_DEFAULT


def parse(argv: list[str] | None) -> argparse.Namespace:
    """
    Process commandline arguments

    :param argv: Command line arguments
    :return: Parsed arguments namespace
    """

    # NOTE: on args with the action 'store_const' with const=True, this is on purpose so as long as the flag is not set,
    # the arg Namespace evaluates to None.

    parser = argparse.ArgumentParser(
        description=START_DESCR,
        prog="censo",
    )

    groups = []

    # RUN SETTINGS
    groups.append(parser.add_argument_group("RUN SETTINGS"))
    groups[0].add_argument(
        "--prescreening",
        "-P",
        dest="prescreening",
        action="store_true",
        help="Enable the prescreening part.",
    )
    groups[0].add_argument(
        "--screening",
        "-S",
        dest="screening",
        action="store_true",
        help="Enable the screening part.",
    )
    groups[0].add_argument(
        "--optimization",
        "-O",
        dest="optimization",
        action="store_true",
        help="Enable the optimization part.",
    )
    groups[0].add_argument(
        "--refinement",
        "-R",
        dest="refinement",
        action="store_true",
        help="Enable the refinement part.",
    )
    groups[0].add_argument(
        "--nmr",
        dest="nmr",
        action="store_true",
        help="Enable the NMR calculation part.",
    )
    groups[0].add_argument(
        "--rot",
        dest="rot",
        action="store_true",
        help="Enable the optical rotation calculation part.",
    )
    groups[0].add_argument(
        "--uvvis",
        dest="uvvis",
        action="store_true",
        help="Enable the UV/Vis calculation part.",
    )
    groups[0].add_argument(
        "-i",
        "--input",
        dest="inp",
        type=str,
        help="Relative path to ensemble file, e.g. crest_conformers.xyz (default).",
        default="crest_conformers.xyz",
    )
    groups[0].add_argument(
        "-n",
        "--nconf",
        dest="nconf",
        type=int,
        help="The first 'nconf' conformers will be considered.",
    )
    groups[0].add_argument(
        "-c",
        "--charge",
        dest="charge",
        default=0,
        type=int,
        help="Integer charge of the investigated molecule.",
    )
    groups[0].add_argument(
        "-u",
        "--unpaired",
        dest="unpaired",
        default=0,
        type=int,
        help="Integer number of unpaired electrons of the investigated molecule.",
    )
    groups[0].add_argument(
        "-v",
        "--version",
        dest="version",
        action="store_true",
        help="Print CENSO version and exit.",
    )
    groups[0].add_argument(
        "--cleanup",
        dest="cleanup",
        action="store_true",
        help="Delete unneeded files from current working directory.",
    )
    groups[0].add_argument(
        "--cleanup-all",
        dest="cleanup_all",
        action="store_true",
        help="Delete all CENSO files from previous runs from current working directory. "
        "Stronger than --cleanup !",
    )
    groups[0].add_argument(
        "--new-config",
        dest="writeconfig",
        action="store_true",
        help="Write new configuration file, which is placed into the current "
        "directory.",
    )
    groups[0].add_argument(
        "--inprc",
        dest="inprcpath",
        help="Use to provide a path to the CENSO configuration file if you want to use a different one"
        " than the default (~/.censo2rc).",
    )
    groups[0].add_argument(
        "--constraints",
        dest="constraints",
        help="Path to a file containing constraints in xtb format for use in geometry optimizations.",
    )
    groups[0].add_argument(
        "--maxcores",
        dest="maxcores",
        type=int,
        help="Number of cores that should be used for CENSO on the machine. If this is not provided CENSO will use "
        "the maximum number available (also checks for slurm environment variables).",
    )
    # groups[0].add_argument(
    #     "-O",
    #     "--omp",
    #     dest="omp",
    #     type=int,
    #     help="Number of OpenMP threads, e.g. 4. Effectively translates to the number of cores used per calculation "
    #     "if load balancing is disabled.",
    # )
    # NOTE: omp arg is not used anymore since the way it works right now, censo will always try to get the minimum number of cores per process
    # because this is the most efficient. therefore, the user should only set the minimum number of cores
    groups[0].add_argument(
        "--omp-min",
        dest="ompmin",
        type=int,
        help=f"Minimum number of OpenMP threads per process, default is {OMPMIN_DEFAULT}. This is mostly important if load balancing is enabled.",
        default=OMPMIN_DEFAULT,
    )
    groups[0].add_argument(
        "--loglevel",
        dest="loglevel",
        help="Set the loglevel for all modules to a specified level.",
        choices=["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"],
        default="INFO",
    )
    groups[0].add_argument(
        "--logpath",
        dest="logpath",
        help="Relative/absolute path to the logfile. If no path is provided, censo.log will be written in the output directory.",
    )
    groups[0].add_argument(
        "--reload",
        dest="reload",
        nargs="+",
        help="Reload data from json output files. List all file names separated by spaces. Will be loaded in order (first to last). "
        "Note that all conformers from the current ensemble need to be included in the output data keys.",
    )
    groups[0].add_argument(
        "--keep-all",
        dest="keep_all",
        action="store_true",
        help="Do not cut down the ensemble, keep all conformers from start to end.",
    )
    groups[0].add_argument(
        "--ignore-failed",
        dest="ignore_failed",
        action="store_true",
        help="Ignore failed calculations. If this is not set, failed calculations will raise RuntimeError.",
    )

    # GENERAL SETTINGS
    groups.append(parser.add_argument_group("GENERAL SETTINGS"))
    groups[1].add_argument(
        "-T",
        "--temperature",
        dest="temperature",
        type=float,
        help="Temperature in Kelvin for thermostatistical evaluation.",
    )
    groups[1].add_argument(
        "--trange",
        dest="trange",
        nargs=3,
        metavar=("start", "end", "step"),
        type=float,
        help="specify a temperature range [start, end, step] e.g.: 250.0 300.0 10.0"
        "  resulting in the range [250.0, 260.0, 270.0, 280.0, 290.0, 300.0].",
    )
    groups[1].add_argument(
        "--bhess",
        dest="bhess",
        action="store_const",
        const=True,
        help="Uses SPH and applies structure constraint to input/DFT geometry "
        "for mRRHO calcuation. ",
    )
    groups[1].add_argument(
        "--consider-sym",
        dest="consider_sym",
        action="store_const",
        const=True,
        help="Consider symmetry in mRRHO calcuation (based on desy xtb threshold). ",
    )
    # groups[1].add_argument(
    #     "--rmsdbias",
    #     dest="rmsdbias",
    #     action="store_const",
    #     const=True,
    #     help="Applies constraint to rmsdpot.xyz to be consistent to CREST. ",
    # )
    groups[1].add_argument(
        "--sm-rrho",
        dest="sm_rrho",
        type=str,
        help="Solvation model used in xTB GmRRHO calculation. Applied if not in "
        "gas-phase. Options are 'gbsa' or 'alpb'.",
    )
    groups[1].add_argument(
        "--evaluate-rrho",
        dest="evaluate_rrho",
        action="store_const",
        const=True,
        help="Evaluate mRRHO contribution.",
    )
    groups[1].add_argument(
        "--solvent",
        dest="solvent",
        type=str,
        help="Solvent to be used for Gsolv calculation.",
    )
    groups[1].add_argument(
        "--gas-phase",
        dest="gas_phase",
        action="store_const",
        const=True,
        help="Run calculation in gas-phase, overriding all solvation settings.",
    )
    groups[1].add_argument(
        "--imagthr",
        dest="imagthr",
        type=float,
        help="threshold for inverting imaginary frequencies for thermo in cm-1,"
        " e.g. -30.0.",
    )
    groups[1].add_argument(
        "--sthr",
        dest="sthr",
        type=float,
        help="Rotor cut-off for thermo in cm-1, e.g. 50.0.",
    )
    groups[1].add_argument(
        "--scale",
        dest="scale",
        type=float,
        help="Scaling factor for frequencies, e.g. 1.0.",
    )
    """
    groups[1].add_argument(
        "--vapor_pressure",
        "-vp",
        dest="vapor_pressure",
        action="store_true",
        help="Gsolv is evaluated for the input molecule in its solution (same). "
        "Only possible with COSMO-RS.",
    )

    # PRESCREENING SETTINGS
    groups.append(parser.add_argument_group("PRESCREENING SETTINGS"))

    # SCREENING SETTINGS
    groups.append(parser.add_argument_group("SCREENING SETTINGS"))

    # OPTIMIZATION SETTINGS
    groups.append(parser.add_argument_group("OPTIMIZATION SETTINGS"))

    # REFINEMENT SETTINGS
    groups.append(parser.add_argument_group("REFINEMENT SETTINGS"))

    # NMR SETTINGS
    groups.append(parser.add_argument_group("NMR SETTINGS"))

    # OPTROT SETTINGS
    groups.append(parser.add_argument_group("OPTROT SETTINGS"))

    # UVVIS SETTINGS
    groups.append(parser.add_argument_group("UVVIS SETTINGS"))

    # leave these options out for now, implementation for cml complicated
    groups[7].add_argument(
        "-freqOR",
        "--freqOR",
        dest="freq_or",
        nargs="*",
        required=False,
        type=float,
        help="Frequencies to evaluate specific rotation at in nm, e.g. 589 "
        "or 589 700 to evaluate at 598 nm and 700 nm.",
    )
    groups[6].add_argument(
        "-couplings",
        "--couplings",
        dest="couplings",
        action="store_true",
        required=False,
        help="Option to run coupling constant calculations. Options are ???.",
    )
    groups[6].add_argument(
        "-shieldings",
        "--shieldings",
        dest="shieldings",
        action="store_true",
        required=False,
        help="Option to run shielding constant calculations. Options are ???.",
    )
    groups[6].add_argument(
        "-hactive",
        "--hactive",
        dest="h_active",
        action="store_true",
        required=False,
        help="Investigates hydrogen nuclei in coupling and shielding calculations."
        "choices=[???]",
    )
    groups[6].add_argument(
        "-cactive",
        "--cactive",
        dest="c_active",
        action="store_true",
        required=False,
        help="Investigates carbon nuclei in coupling and shielding calculations."
        "choices=[???]",
    )
    groups[6].add_argument(
        "-factive",
        "--factive",
        dest="f_active",
        action="store_true",
        required=False,
        help="Investigates fluorine nuclei in coupling and shielding calculations."
        "choices=[???]",
    )
    groups[6].add_argument(
        "-siactive",
        "--siactive",
        dest="si_active",
        action="store_true",
        required=False,
        help="Investigates silicon nuclei in coupling and shielding calculations."
        "choices=[???]",
    )
    groups[6].add_argument(
        "-pactive",
        "--pactive",
        dest="p_active",
        action="store_true",
        required=False,
        help="Investigates phosophorus nuclei in coupling and shielding calculations."
        "choices=[???]",
    )
    groups[4].add_argument(
        "-crestcheck",
        "--crestcheck",
        dest="crestcheck",
        action="store_true",
        required=False,

        help="Option to sort out conformers after DFT ensembleopt which CREST "
        "identifies as identical or rotamers of each other. \nThe identification/"
        "analysis is always performed, but the removal of conformers has to "
        "be the choice of the user. Options are: [???]", # TODO
    )
    groups[4].add_argument(
        "-macro",
        dest="macrocycles",
        action="store_const",
        const=True,
        required=False,
        help="Option to use macrocycles for geometry optimization."
    )
    groups[4].add_argument(
        "-optlevel2",
        "--optlevel2",
        dest="optlevel2",
        default=None,
        required=False,
        help="Option to set the optlevel in part2, only if optimizing with the xTB-optimizer!"
        "Allowed values are ***", # TODO
    )
    groups[4].add_argument(
        "-optcycles",
        "--optcycles",
        dest="optcycles",
        required=False,
        type=int,
        help="number of cycles in ensemble optimizer.",
    )
    groups[4].add_argument(
        "-hlow",
        "--hlow",
        dest="hlow",
        required=False,
        type=float,
        help="Lowest force constant in ANC generation (real), used by xTB-optimizer.",
    )
    groups[4].add_argument(
        "-spearmanthr",
        "--spearmanthr",
        dest="spearmanthr",
        required=False,
        help="Value between -1 and 1 for the spearman correlation coeffient threshold, "
        "employed in the ensemlbe optimizer",
    )
    groups[4].add_argument(
        "-radsize",
        "--radsize",
        dest="radsize",
        required=False,
        type=int,
        help="Radsize used in the ensembleopt and only for r2scan-3c!",
    )
    group1.add_argument(
        "-func",
        "--functional",
        dest="func",
        choices=options.value_options["func"],
        action="store",
        required=False,

        help="Functional for geometry ensembleopt (used in part2) and "
        "single-points in part1",
    )
    group1.add_argument(
        "-basis",
        "--basis",
        dest="basis",
        action="store",
        required=False,

        help="Basis set employed together with the functional (func) for the "
        "low level single point in part1 und ensembleopt in part2.",
    )
    group1.add_argument(
        "-prog",
        "--prog",
        choices=options.value_options["prog"],
        dest="prog",
        required=False,

        help="QM-program used in part0, part1 and part2 either 'orca' or 'tm'.",
    )
    group10.add_argument(
        "-part0_gfnv",
        "--part0_gfnv",
        dest="part0_gfnv",
        choices=options.value_options["part0_gfnv"],

        action="store",
        required=False,
        help="GFNn-xTB version employed for calculating the GFNn-xTB "
        "single point in part0. "
        f"Allowed values are [{', '.join(options.value_options['part0_gfnv'])}]",
    )
    group3.add_argument(
        "-part1",
        "--part1",
        choices=["on", "off"],
        dest="part1",
        action="store",
        required=False,

        help="Option to turn the prescreening evaluation (part1) 'on' or 'off'.",
    )
    group3.add_argument(
        "-smgsolv1",
        "--smgsolv1",
        choices=options.value_options["smgsolv1"],
        dest="smgsolv1",
        action="store",
        required=False,

        help="Solvent model for the Gsolv evaluation in part1. This can either be"
        " an implicit solvation or an additive solvation model. "
        f"Allowed values are [{', '.join(options.value_options['smgsolv1'])}]",
    )
    group10.add_argument(
        "-prescreening_threshold",
        "-prethr",
        "--thresholdpre",
        dest="prescreening_threshold",

        action="store",
        type=float,
        required=False,
        help=(
            "Threshold in kcal/mol. All conformers in part0 (prescreening)"
            " with a relativ energy below the threshold are considered for part1."
        ),
    )
    group4.add_argument(
        "-sm2",
        "--solventmodel2",
        choices=options.value_options.get("sm2"),
        dest="sm2",
        action="store",
        required=False,

        help="Solvent model employed during the geometry ensembleopt in part2."
        "The solvent model sm2 is not used for Gsolv evaluation, but for the "
        "implicit effect on a property (e.g. the geometry in the ensembleopt).",
    )
    group4.add_argument(
        "-smgsolv2",
        "--smgsolv2",
        choices=options.value_options["smgsolv2"],
        dest="smgsolv2",
        action="store",
        required=False,

        help="Solvent model for the Gsolv (solvation contribution to free energy) "
        "calculation in part2. Either the solvent"
        " model of the ensembleopt (sm2) or an additive solvation model. "
        f"Allowed values are [{', '.join(options.value_options['smgsolv2'])}]",
    ) """

    # TODO - keep this?
    """ group1.add_argument(
        "-prog_rrho",
        "--prog_rrho",
        choices=options.value_options["prog_rrho"],
        dest="prog_rrho",
        required=False,

        help="QM-program for mRRHO contribution in part1 2 and 3, currently only 'xtb'.",
    ) """

    # TODO - keep?
    """ group4.add_argument(
        "-ancopt",
        choices=["on"],  # there is no other option right now!
        dest="ancopt",
        required=False,

        help="Option to use xtb as driver for the xTB-optimizer in part2. "
        "Which is currently not changeable!",
    ) """

    args = parser.parse_args(argv)
    # NOTE: as opposed to previous behaviour '-i' and '--maxcores' are now optional,
    # however the program WILL exit after creating a new config file if any of
    # '--new-config', '--cleanup', '--cleanup-all'
    # are used

    return args



import pytest
from pathlib import Path
import shutil

from censo.params import XtbSolvMod, TmSolvMod, OrcaSolvMod, Prog
from censo.config.parts_config import PartsConfig
from censo.parallel import get_cluster


@pytest.fixture
def fixtures_path():
    return Path(__file__).parent / "fixtures"


@pytest.fixture
def example_ensemble_file(tmp_path: Path):
    """Copies the fixture ensemble file to the test dir and yield the path to that copy."""
    src = Path(__file__).parent / "fixtures" / "crest_conformers.xyz"
    dst = tmp_path / src.name
    shutil.copy(src, dst)
    return dst


# @pytest.fixture(autouse=True)
# def tmp_wd(tmp_path, monkeypatch):
#     orig = os.getcwd()
#     monkeypatch.chdir(tmp_path)
#     yield
#     os.chdir(orig)


@pytest.fixture
def mock_functionals():
    """Mock functional definitions for testing"""
    return {
        "pbe-d4": {
            Prog.TM.value: "actual_value",
            Prog.ORCA.value: "actual_value",
            "disp": "d4",
            "type": "gga",
        },
        "r2scan-3c": {
            Prog.TM.value: "actual_value",
            Prog.ORCA.value: "actual_value",
            "disp": "3c",
            "type": "mgga",
        },
    }


@pytest.fixture
def mock_solvents():
    """Mock solvent definitions for testing"""
    return {
        "h2o": {
            XtbSolvMod.GBSA: "actual_value",
            TmSolvMod.COSMORS: "actual_value",
            OrcaSolvMod.CPCM: "actual_value",
        },
        "chcl3": {
            XtbSolvMod.GBSA: "actual_value",
            TmSolvMod.COSMORS: "actual_value",
            OrcaSolvMod.CPCM: "actual_value",
        },
    }


# Patch PartsConfig.model_validate to default context['check_paths'] = False unless explicitly set
@pytest.fixture(autouse=True)
def patch_model_validate(monkeypatch):
    original_model_validate = PartsConfig.model_validate

    def patched_model_validate(self, *args, **kwargs):
        # Support both positional and keyword context
        context = kwargs.get("context")
        if context is None and len(args) > 0:
            context = args[0]
        if context is None:
            context = {}
        # If check_paths is not set, default to False
        if "check_paths" not in context:
            context["check_paths"] = False
        # Ensure context is passed as a kwarg for consistency
        kwargs["context"] = context
        return original_model_validate(self, **kwargs)

    monkeypatch.setattr(PartsConfig, "model_validate", patched_model_validate)


@pytest.fixture(scope="session")
def parallel_setup():
    """Provide real parallel setup for tests that need it."""

    cluster = get_cluster()
    client = cluster.get_client()
    return client, cluster



import os
import sys
from argparse import ArgumentError, Namespace
from typing import cast, TYPE_CHECKING
from pathlib import Path
import time
from time import sleep

from pydantic import ValidationError

if TYPE_CHECKING:
    from ..ensemble import EnsembleData
    from ..config import PartsConfig

from ..params import DESCR, __version__, Returncode, PLENGTH
from ..utilities import printf
from ..logging import setup_logger, set_loglevel
from .cml_parser import parse
from ..utilities import print_validation_errors, get_time

logger = setup_logger(__name__)


def entry_point(argv: list[str] | None = None) -> Returncode:
    """
    Console entry point to execute CENSO from the command line.

    :param argv: Command line arguments
    :return: Return code
    """
    try:
        args = parse(argv)
    except ArgumentError as e:
        printf(e.message)
        return Returncode.ARGUMENT_ERROR
    except SystemExit as e:
        return cast(Returncode, e.code)

    if not any(vars(args).values()):
        printf("CENSO needs at least one argument!")
        return Returncode.ARGUMENT_ERROR

    # Startup description
    print(DESCR + "\n")

    # Print program call
    printf("CALL: " + " ".join(arg for arg in sys.argv))

    from ..ensembleopt import prescreening, screening, optimization, refinement
    from ..properties import nmr, uvvis, rot

    tasks = [
        ("prescreening", getattr(args, "prescreening", False), prescreening),
        ("screening", getattr(args, "screening", False), screening),
        ("optimization", getattr(args, "optimization", False), optimization),
        ("refinement", getattr(args, "refinement", False), refinement),
        ("nmr", getattr(args, "nmr", False), nmr),
        ("rot", getattr(args, "rot", False), rot),
        ("uvvis", getattr(args, "uvvis", False), uvvis),
    ]

    context = {"check": [name for name, enabled, _ in tasks if enabled]}

    try:
        ensemble, parts_config = startup(args, context)
    except SystemExit as e:
        return cast(Returncode, e.code)

    # Set up parallel managers once for the entire run
    from ..parallel import get_cluster

    try:
        cluster = get_cluster(maxcores=args.maxcores, ompmin=args.ompmin)
        client = cluster.get_client()
    except Exception as e:
        printf(f"Encountered error in setting up parallelization: {e}")
        return Returncode.GENERIC_ERROR

    comparison: dict[str, dict[str, float]] = {}
    cut: bool = not args.keep_all
    times: dict[str, float] = {}

    logger.debug("Setting up parallel managers for CLI run...")
    try:
        for partname, func in [
            (partname, task) for partname, enabled, task in tasks[:4] if enabled
        ]:
            start = time.perf_counter()
            _ = func(ensemble, parts_config, client, cut=cut)  # type: ignore[operator]
            end = time.perf_counter()
            runtime = end - start
            printf(f"Ran {func.__name__} in {runtime:.2f} seconds!")
            times[partname] = runtime

            # Collect results for comparison
            from ..params import AU2KCAL

            mingtot = min(conf.gtot for conf in ensemble)
            comparison[func.__name__] = {
                conf.name: (conf.gtot - mingtot) * AU2KCAL for conf in ensemble
            }

        if len(comparison) > 0:
            print("\nFinished ensemble optimization\n")

            # Print final comparison
            print_comparison(comparison)

        if any(enabled for _, enabled, _ in tasks[4:]):
            print("\nRunning property calculations\n")

        for partname, func in [
            (partname, task) for partname, enabled, task in tasks[4:] if enabled
        ]:
            start = time.perf_counter()
            _ = func(ensemble, parts_config, client)  # type: ignore[operator]
            end = time.perf_counter()
            runtime = end - start
            printf(f"Ran {func.__name__} in {runtime:.2f} seconds!")
            times[partname] = runtime
    except Exception:
        import traceback

        tb = traceback.format_exc()
        logger.debug(f"Encountered exception:\n{tb}")

        # Save as much data as possible
        printf(
            "Encountered exception. Stopping CENSO and dumping most recent ensemble."
        )
        ensemble.dump_json(Path("CRASH_DUMP.json"))
        ensemble.dump_xyz(Path("CRASH_DUMP.xyz"))
        sleep(5)  # Make dask exit more graceful
        return Returncode.GENERIC_ERROR

    total_time = sum(times.values())

    printf("\nTimings:")
    for part in times:
        seconds, minutes, hours = get_time(times[part])
        printf(
            f"{part.capitalize():>19}: {hours:02d}:{minutes:02d}:{seconds:02d}  ({times[part] / total_time * 100:5.1f} %)"
        )

    printf("=" * 40)
    seconds, minutes, hours = get_time(total_time)
    printf(f"\n{'Total CENSO runtime':>19}: {hours:02d}:{minutes:02d}:{seconds:02d}")

    printf("\nCENSO all done!")
    sleep(5)  # Make dask exit more graceful
    return Returncode.OK


# sets up a ensemble object using the given cml arguments and censorc
def startup(
    args: Namespace, context: dict[str, list[str]]
) -> tuple["EnsembleData", "PartsConfig"]:
    """
    Set up the ensemble and configuration using command line arguments and context.

    :param args: Parsed command line arguments
    :param context: Context dictionary with enabled tasks
    :return: Tuple of ensemble data and parts configuration
    """
    from logging import DEBUG

    from ..config.setup import configure, write_rcfile
    from ..ensemble import EnsembleData
    from ..logging import set_filehandler

    # Load up the ensemble and configure settings
    cwd = os.getcwd()

    # run actions for which no complete setup is needed
    if args.version:
        printf(__version__)
        sys.exit(Returncode.OK)
    elif args.cleanup:
        cleanup_run(cwd)
        printf("Removed files and going to exit!")
        sys.exit(Returncode.OK)
    elif args.cleanup_all:
        cleanup_run(cwd, complete=True)
        printf("Removed files and going to exit!")
        sys.exit(Returncode.OK)
    elif args.writeconfig:
        write_rcfile(Path() / "censo2rc_NEW")
        sys.exit(Returncode.OK)

    if len(context.get("check", [])) == 0:
        printf("No tasks enabled! Stopping CENSO.")
        sys.exit(Returncode.OK)

    try:
        parts_config = configure(rcpath=args.inprcpath, args=args, context=context)
    except ValidationError as e:
        print_validation_errors(e)
        sys.exit(Returncode.CONFIG_ERROR)

    # Set up logging
    set_loglevel(args.loglevel)
    logpath = Path(args.logpath or Path(args.inp).parent / "censo.log").resolve()
    set_filehandler(logpath)

    # initialize ensemble
    ensemble = EnsembleData()

    # read input and setup conformers
    try:
        ensemble.read_input(args.inp, args.charge or 0, args.unpaired or 0, args.nconf)
    except FileNotFoundError:
        printf(f"Could not find input file {Path(args.inp).resolve()}.")
        sys.exit(Returncode.INPUT_NOT_FOUND)

    if args.constraints:
        try:
            ensemble.constraints = Path(args.constraints).read_text()
        except FileNotFoundError:
            printf(
                f"Could not find constraints file {Path(args.constraints).resolve()}."
            )
            sys.exit(Returncode.CONSTRAINTS_NOT_FOUND)

    # if data should be reloaded, do it here
    if args.reload:
        for filename in args.reload:
            ensemble.read_output(Path(filename))

    # Set working directory to input parent
    # NOTE: this is the most convenient way to solve this for CLI version
    wd = Path(args.inp).parent.resolve()
    printf(f"Setting working directory to {wd}")
    os.chdir(wd)

    # Print all active parts settings once
    if logger.level == DEBUG:
        for _, config in parts_config:
            # Print
            printf(config)
    else:
        for fieldname, config in parts_config:
            if (
                getattr(args, fieldname, False)
                or fieldname == "general"
                or fieldname == "paths"
            ):
                # Print
                printf(config)

    printf("\n" + "".ljust(int(PLENGTH), "-") + "\n")

    # END of setup
    # -> ensemble.conformers contains all conformers with their info from input (sorted by CREST energy if possible)
    # -> settings are updated with cml args
    # -> output data is reloaded if wanted
    # -> working directory is set

    return ensemble, parts_config


def cleanup_run(cwd: str | Path, complete: bool = False):
    """
    Delete all unneeded files.

    :param cwd: Current working directory
    :param complete: If true, delete all files including ensembles and outputs
    """
    import shutil

    # files containing these patterns are deleted
    to_delete = [
        "censo.log",
        "0_PRESCREENING",
        "1_SCREENING",
        "2_OPTIMIZATION",
        "3_REFINEMENT",
        "4_NMR",
        "6_UVVIS",
    ]

    if complete:
        printf(
            "Removing ALL files generated by previous CENSO runs, including ensembles and output files (*.xyz, *.out, *.json)!"
        )

    printf(
        f"Be aware that files in {cwd} and subdirectories with names containing the following substrings "
        f"will be deleted:"
    )
    for sub in to_delete:
        printf(sub)

    printf("Do you wish to continue?")
    printf("Please type 'yes'/'y' or 'no'/'n':")

    ui = input()

    while ui.strip().lower() not in ["yes", "y", "no", "n"]:
        printf("Please type 'yes'/'y' or 'no'/'n':")
        ui = input()

    if ui.strip().lower() in ["no", "n"]:
        printf("Aborting cleanup!")
        sys.exit(0)

    # iterate over files in cwd and subdirs recursively and remove them if to delete
    for subdir, _, files in os.walk(cwd):
        if any(s in subdir for s in to_delete):
            printf(f"Removing: {subdir}")
            shutil.rmtree(subdir)
        for file in files:
            if any(s in file for s in to_delete) and (
                complete
                or all(
                    not file.endswith(pattern) for pattern in [".xyz", ".out", ".json"]
                )
            ):
                printf(f"Removing: {file}")
                os.remove(os.path.join(subdir, file))


def print_comparison(comparison: dict[str, dict[str, float]]):
    """
    Print a comparison table of the final rankings.

    :param comparison: Dictionary of comparison data
    """
    from tabulate import tabulate
    from ..utilities import h1

    if len(comparison) > 1:
        printf(h1("FINAL RANKING COMPARISON"))

        headers = ["CONF#"]

        headers.extend([f"ΔGtot {part.capitalize()}" for part in comparison])

        units = [
            "",
        ]

        units.extend(["[kcal/mol]" for _ in headers[1:]])

        confs = [conf for conf in list(comparison.values())[-1]]

        printmap = {"CONF#": lambda confname: confname}

        def make_printmap(h: str):
            return lambda confname: f"{comparison[h.split()[1].lower()][confname]:.2f}"

        for header in headers[1:]:
            printmap[header] = make_printmap(header)

        rows = [
            [printmap[header](confname) for header in headers] for confname in confs
        ]

        for i in range(len(headers)):
            headers[i] += "\n" + units[i]

        table = tabulate(
            rows,
            headers=headers,
            colalign=["center" for _ in headers],
            disable_numparse=True,
            numalign="decimal",
        )
        print(table, flush=True)

        printf("".ljust(int(PLENGTH), "-") + "\n")



import pytest
from censo.config.parts.rot import RotConfig
from censo.params import QmProg


def test_rot_config_default_values():
    """Test default values of RotConfig"""
    config = RotConfig()

    assert config.prog == QmProg.TM
    assert config.func == "pbe-d4"
    assert config.basis == "def2-svpd"
    assert config.freq == [589.0, 633.0]
    assert config.template is False


@pytest.mark.parametrize(
    "freq,should_pass",
    [
        ([589.0], True),
        ([589.0, 633.0, 1000.0], True),
        ([], False),
        ([0.0], False),
        ([-1.0], False),
        ([589.0, -1.0], False),
    ],
)
def test_freq_validation(freq, should_pass):
    """Test frequency validation"""
    if should_pass:
        config = RotConfig(freq=freq)
        assert config.freq == freq
    else:
        with pytest.raises(ValueError):
            RotConfig(freq=freq)


@pytest.mark.parametrize(
    "freq_str,expected",
    [
        ("[589.0]", [589.0]),
        ("[589.0, 633.0]", [589.0, 633.0]),
    ],
)
def test_freq_casting(freq_str, expected):
    """Test frequency string casting"""
    config = RotConfig(freq=freq_str)  # type: ignore[arg-type]
    assert config.freq == expected


def test_invalid_freq_casting():
    """Test invalid frequency string casting"""
    with pytest.raises(ValueError):
        RotConfig(freq="invalid")  # type: ignore[arg-type]


@pytest.mark.parametrize(
    "func,should_pass",
    [
        ("pbe-d4", True),
        ("r2scan-3c", True),
        ("invalid-func", False),
    ],
)
def test_functional_validation(func, should_pass):
    """Test functional validation for TM program"""
    if should_pass:
        config = RotConfig(func=func)
        assert config.func == func
    else:
        with pytest.raises(ValueError):
            RotConfig(func=func)



import pytest
from censo.config.parts.optimization import OptimizationConfig
from censo.params import QmProg, GfnVersion, TmSolvMod, OrcaSolvMod


def test_optimization_config_default_values():
    """Test default values of OptimizationConfig"""
    config = OptimizationConfig()

    assert config.prog == QmProg.TM
    assert config.func == "r2scan-3c"
    assert config.basis == "def2-mtzvpp"
    assert config.sm == TmSolvMod.DCOSMORS
    assert config.gfnv == GfnVersion.GFN2
    assert config.optcycles == 8
    assert config.maxcyc == 200
    assert config.optlevel == "normal"
    assert config.threshold == 3.0
    assert config.gradthr == 0.01
    assert config.hlow == 0.01
    assert config.macrocycles is True
    assert config.constrain is False
    assert config.xtb_opt is True
    assert config.template is False


def test_positive_integer_validation():
    """Test validation of positive integer parameters"""
    with pytest.raises(ValueError):
        OptimizationConfig(optcycles=0)
    with pytest.raises(ValueError):
        OptimizationConfig(optcycles=-1)
    with pytest.raises(ValueError):
        OptimizationConfig(maxcyc=0)
    with pytest.raises(ValueError):
        OptimizationConfig(maxcyc=-1)


def test_positive_float_validation():
    """Test validation of positive float parameters"""
    with pytest.raises(ValueError):
        OptimizationConfig(threshold=0)
    with pytest.raises(ValueError):
        OptimizationConfig(threshold=-1.0)
    with pytest.raises(ValueError):
        OptimizationConfig(gradthr=0)
    with pytest.raises(ValueError):
        OptimizationConfig(gradthr=-0.01)
    with pytest.raises(ValueError):
        OptimizationConfig(hlow=0)
    with pytest.raises(ValueError):
        OptimizationConfig(hlow=-0.01)


@pytest.mark.parametrize(
    "optlevel",
    ["crude", "sloppy", "loose", "lax", "normal", "tight", "vtight", "extreme"],
)
def test_valid_optlevels(optlevel):
    """Test all valid optimization levels"""
    config = OptimizationConfig(optlevel=optlevel)
    assert config.optlevel == optlevel


def test_invalid_optlevel():
    """Test invalid optimization level"""
    with pytest.raises(ValueError):
        OptimizationConfig(optlevel="invalid")  # type: ignore[arg-type]


@pytest.mark.parametrize(
    "sm", [TmSolvMod.DCOSMORS, TmSolvMod.COSMO, OrcaSolvMod.CPCM, OrcaSolvMod.SMD]
)
def test_valid_solvent_models(sm):
    """Test valid solvent model combinations"""
    config = OptimizationConfig(sm=sm)
    assert config.sm == sm


def test_invalid_solvent_model():
    """Test invalid solvent model"""
    with pytest.raises(ValueError):
        OptimizationConfig(sm=TmSolvMod.COSMORS)  # type: ignore[arg-type]


@pytest.mark.parametrize(
    "prog,func,should_pass",
    [
        (QmProg.TM, "r2scan-3c", True),
        (QmProg.ORCA, "r2scan-3c", True),
        (QmProg.TM, "invalid-func", False),
        (QmProg.ORCA, "invalid-func", False),
    ],
)
def test_functional_validation(prog, func, should_pass):
    """Test functional validation with different program combinations"""
    if should_pass:
        config = OptimizationConfig(prog=prog, func=func)
        assert config.prog == prog
        assert config.func == func
    else:
        with pytest.raises(ValueError):
            OptimizationConfig(prog=prog, func=func)


def test_constraints_validation():
    """Test constraints validation"""
    # Valid: constraints with xtb_opt enabled
    config = OptimizationConfig(constrain=True, xtb_opt=True)
    assert config.constrain is True
    assert config.xtb_opt is True

    # Invalid: constraints without xtb_opt
    with pytest.raises(
        ValueError, match="Constraints can currently only be used with ANCOPT"
    ):
        OptimizationConfig(constrain=True, xtb_opt=False)



import pytest
from censo.ensembleopt.screening import screening
from censo.ensemble import EnsembleData
from censo.config.parts_config import PartsConfig
from censo.params import SOLV_MODS, OrcaSolvMod, QmProg, TmSolvMod
from dask.distributed import Client


# Parameterized test cases for screening
@pytest.mark.optional
@pytest.mark.parametrize(
    "prog, solvation_model, evaluate_rrho",
    [
        pytest.param(QmProg.ORCA, None, True, marks=pytest.mark.requires_orca),
        pytest.param(QmProg.TM, None, True, marks=pytest.mark.requires_turbomole),
        pytest.param(
            QmProg.ORCA, OrcaSolvMod.SMD, True, marks=pytest.mark.requires_orca
        ),
        pytest.param(
            QmProg.TM, TmSolvMod.DCOSMORS, True, marks=pytest.mark.requires_turbomole
        ),
        *[  # type: ignore[var-annotated]
            pytest.param(QmProg.ORCA, sm, False, marks=pytest.mark.requires_orca)
            for sm in SOLV_MODS[QmProg.ORCA.value]
        ],
        *[  # type: ignore[var-annotated]
            pytest.param(
                QmProg.TM,
                sm,
                False,
                marks=(
                    (pytest.mark.requires_turbomole, pytest.mark.requires_cosmotherm)
                    if sm in [TmSolvMod.COSMORS, TmSolvMod.COSMORS_FINE]
                    else pytest.mark.requires_turbomole
                ),
            )
            for sm in SOLV_MODS[QmProg.TM.value]
        ],
    ],
)
def test_screening_integration(
    config: PartsConfig,
    ensemble_from_xyz: EnsembleData,
    client: Client,
    prog: QmProg,
    solvation_model: TmSolvMod | OrcaSolvMod | None,
    evaluate_rrho: bool,
):
    config.general.evaluate_rrho = evaluate_rrho
    config.screening.prog = prog
    if solvation_model is not None:
        config.general.gas_phase = False
        config.screening.sm = solvation_model
        config.general.solvent = "h2o"
        # For COSMORS also test the case where the user might have forgotten to set gsolv_included to False
        if solvation_model == TmSolvMod.COSMORS:
            config.screening.gsolv_included = True
    else:
        config.general.gas_phase = True
    result = screening(ensemble_from_xyz, config, client, cut=True)
    assert result is not None
    # Optionally add more output checks



"""Tests for GeneralConfig"""

import pytest
from censo.config.parts.general import GeneralConfig
from censo.params import XtbSolvMod


def test_general_config_default_values():
    """Test default values of GeneralConfig"""
    config = GeneralConfig()

    assert config.temperature == 298.15
    # assert config.multitemp is True
    # assert config.trange == (273.15, 373.15, 5)
    assert config.evaluate_rrho is True
    assert config.sm_rrho == XtbSolvMod.GBSA
    # assert config.consider_sym is True
    # assert config.bhess is True
    # assert config.rmsdbias is False
    assert config.imagthr == -100.0
    assert config.sthr == 50.0
    assert config.solvent == "h2o"
    assert config.gas_phase is False
    assert config.copy_mo is True
    assert config.balance is True


def test_temperature_validation():
    """Test temperature validation"""
    with pytest.raises(ValueError):
        GeneralConfig(temperature=-1.0)


# def test_trange_validation():
#     """Test trange validation"""
#     # Test negative temperature
#     with pytest.raises(ValueError, match="Negative temperature"):
#         GeneralConfig(trange=(-1.0, 373.15, 5))
#
#     # Test wrong order
#     with pytest.raises(ValueError, match="must be larger than"):
#         GeneralConfig(trange=(373.15, 273.15, 5))
#
#     # Test invalid step size
#     with pytest.raises(ValueError, match="must be positive"):
#         GeneralConfig(trange=(273.15, 373.15, 0))
#     with pytest.raises(ValueError, match="must be positive"):
#         GeneralConfig(trange=(273.15, 373.15, -1))


def test_imagthr_validation():
    """Test imagthr validation"""
    with pytest.raises(ValueError):
        GeneralConfig(imagthr=0.0)


@pytest.mark.parametrize(
    "solvent,sm_rrho,should_pass",
    [
        ("h2o", XtbSolvMod.GBSA, True),
        ("invalid_solvent", XtbSolvMod.GBSA, False),
    ],
)
def test_solvent_validation(solvent, sm_rrho, should_pass):
    """Test solvent validation with different combinations"""
    if should_pass:
        config = GeneralConfig(solvent=solvent, sm_rrho=sm_rrho)
        assert config.solvent == solvent
        assert config.sm_rrho == sm_rrho
    else:
        with pytest.raises(ValueError):
            GeneralConfig(solvent=solvent, sm_rrho=sm_rrho)



import os
import shutil
from argparse import Namespace
from typing import Any
from pathlib import Path
from configparser import ConfigParser
import warnings


from ..config.paths import PathsConfig
from .parts_config import PartsConfig
from ..params import CENSORCNAME, QmProg
from ..logging import setup_logger

logger = setup_logger(__name__)


def configure(
    rcpath: str | None = None,
    args: Namespace | None = None,
    context: dict[str, list[str]] | None = None,
) -> PartsConfig:
    """
    Configure the application based on the provided configuration file path.

    If no configuration file path is provided, it searches for the default configuration file.
    If no configuration file is found, it raises a FileNotFoundError.

    :param rcpath: Path to the configuration file.
    :param args: Parsed command line arguments. Defaults to None.
    :param context: Additional context for validation. Defaults to None.
    :return: Configuration instance.
    """
    # Try to find the .censo2rc in the user's home directory
    # if no configuration file path is provided
    # NOTE: order of priority for rcfile path:
    # - args.inprcfile
    # - home dir
    # - env variable
    if rcpath is None:
        censorc_path = find_rcfile()
    else:
        if not Path(rcpath).resolve().is_file():
            raise FileNotFoundError(f"No configuration file found at {rcpath}.")
        censorc_path = Path(rcpath).resolve()

    if censorc_path is not None:
        # Read the actual configuration file (located at rcpath if not None, otherwise rcfile in home dir)
        settings_dict = read_rcfile(censorc_path, silent=False)

        # Create configurations without solvlent and paths validation for now (will be validated in the end after cml args)
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore")
            parts_config = PartsConfig.model_validate(settings_dict)
    else:
        # Try to find paths
        paths = find_program_paths()

        # Create default configurations with auto-detected paths (no path or solvent validation)
        parts_config = PartsConfig.model_validate({"paths": paths})

    # Override general settings only for now
    for field in parts_config.general.__class__.model_fields:
        setting = getattr(args, field, None)
        if setting is not None:
            parts_config.general.__setattr__(field, setting)

    # Revalidate after command line args incl. solvents and paths with context
    parts_config = PartsConfig.model_validate(parts_config, context=context)

    return parts_config


def read_rcfile(path: Path, silent: bool = True) -> dict[str, dict[str, Any]]:
    """
    Read the configuration file at 'path' and return the settings as a dictionary.

    :param path: Path to the configuration file.
    :param silent: If True, no messages will be printed.
    :return: Dictionary containing the settings read from the configuration file.
    """
    # read config file
    if not silent:
        print(f"Reading configuration file from {path}.")

    parser = ConfigParser()
    parser.read_string(path.read_text())

    return {section: dict(parser[section]) for section in parser.sections()}


def write_rcfile(path: Path) -> None:
    """
    Write new configuration file with default settings into file at 'path'.

    Also reads program paths from preexisting configuration file or tries to
    determine the paths automatically.

    :param path: Path to the new configuration file.
    :return: None
    """
    # what to do if there is an existing configuration file
    paths = None
    if path.is_file():
        print(
            f"An existing configuration file has been found at {path}.\n",
            f"Renaming existing file to {path.name}_OLD.\n",
        )
        # Read program paths from the existing configuration file
        print("Reading program paths from existing configuration file ...")
        paths = read_program_paths(path)

        # Rename existing file
        path.rename(f"{path}_OLD")

    with open(path, "w", newline=None) as rcfile:
        parser = ConfigParser()

        # Try to get paths from 'which'
        print("Trying to determine program paths automatically ...")
        paths = find_program_paths()

        # collect all default settings from parts and feed them into the parser
        # Mute warnings
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            parts_config = PartsConfig.model_validate({"paths": paths})

        parser.read_dict(parts_config.model_dump(mode="json"))

        print(f"Writing new configuration file to {path} ...")
        parser.write(rcfile)

    print(
        f"\nA new configuration file was written into {path}.\n"
        + "You should adjust the settings to your needs and set the program paths.\n"
        + "Right now the settings are at their default values.\n"
    )

    if CENSORCNAME not in path.stem:
        print(
            f"To load it automatically make sure that the file name is '{CENSORCNAME}' and it's located in your home directory.\n"
            f"Current name: '{os.path.split(path)[-1]}'.\n"
        )


def read_program_paths(path: Path) -> dict[str, str]:
    """
    Read program paths from the configuration file at 'path'.

    :param path: Path to the configuration file.
    :return: Dictionary of program paths.
    """
    parser = ConfigParser()
    parser.read_string(path.read_text())
    paths = parser.has_section("paths")

    if not paths:
        logger.warning(f"No paths found in {path}")
        return dict()

    return dict(parser["paths"])


def find_program_paths() -> dict[str, str]:
    """
    Try to determine program paths automatically.

    :return: Dictionary of found program paths.
    """
    paths: dict[str, str] = {}
    for program in PathsConfig.model_fields.keys():
        path = shutil.which(program)

        if path:
            paths[program] = path

    if QmProg.TM.value in PathsConfig.model_fields.keys():
        path = shutil.which("ridft")
        if path:
            paths[QmProg.TM.value] = str(Path(path).parent)

    # If cosmotherm is found try to set cosmorssetup automatically
    if "cosmotherm" in paths:
        # cosmotherm path parent should be BIN-LINUX, CTDATA-FILES is on the same level
        ctdata = (Path(paths["cosmotherm"]).parent / ".." / "CTDATA-FILES").resolve()
        for file in ctdata.glob("*.ctd"):
            if "BP_TZVP" in file.name and file.is_file():
                paths["cosmorssetup"] = file.name

    return paths


def find_rcfile() -> Path | None:
    """
    Check for existing .censorc2 in $home dir or rcfile path in environment variable.

    :return: Path to the configuration file if found, else None.
    """

    rcpath = None
    homepath = Path("~").expanduser() / CENSORCNAME
    try:
        envpath = (
            Path(os.environ["CENSORC_PATH"]) if "CENSORC_PATH" in os.environ else None
        )
    except TypeError:
        envpath = None

    # check for .censorc in $home
    if homepath.is_file():
        rcpath = homepath
    # check for path in env variable
    elif envpath is not None and envpath.is_file():
        rcpath = envpath

    return rcpath



from typing import Literal, Self
from pydantic import model_validator, Field

from .base import BasePartConfig
from ...assets import FUNCTIONALS
from ...params import OrcaSolvMod, QmProg


class UVVisConfig(BasePartConfig):
    """Config class for UVVis"""

    prog: Literal[QmProg.ORCA] = QmProg.ORCA
    """Program that should be used for calculations."""

    func: str = "wb97x-d4"
    """Functional that should be used for calculations."""

    basis: str = "def2-TZVP"
    """Basis set that should be used for calculations."""

    sm: OrcaSolvMod = OrcaSolvMod.SMD
    """Solvation model that should be used for calculations."""

    nroots: int = Field(gt=0, default=20)
    """Number of roots to calculate."""

    template: bool = False
    """Whether to use template files."""

    @model_validator(mode="after")
    def func_must_be_known_in_prog(self) -> Self:
        """
        Validate that the functional is known for the chosen program.

        :return: The validated instance.
        """
        prog: str = self.prog
        try:
            assert FUNCTIONALS[self.func][prog] is not None
            assert FUNCTIONALS[self.func]["disp"] is not None
            assert FUNCTIONALS[self.func]["type"] is not None
        except (KeyError, AssertionError):
            raise ValueError(
                f"Functional {self.func} not (fully) defined for prog {prog}."
            )
        return self



from tabulate import tabulate
from pathlib import Path
from typing import Any
import json
from collections.abc import Callable
from dask.distributed import Client

from ..molecules import MoleculeData, Contributions
from ..logging import setup_logger
from ..parallel import execute
from ..params import AU2KCAL, PLENGTH, GridLevel, Prog
from ..utilities import h1, h2, printf, Factory, DataDump
from ..config import PartsConfig, RefinementConfig
from ..config.job_config import RRHOJobConfig, SPJobConfig
from ..ensemble import EnsembleData
from ..processing import QmProc, XtbProc

logger = setup_logger(__name__)


# TODO: this is not very DRY considering the screening function
def refinement(
    ensemble: EnsembleData,
    config: PartsConfig,
    client: Client,
    cut: bool = True,
) -> dict[str, Any]:
    """
    Basically the same as screening, however here we use a Boltzmann population cutoff instead of kcal cutoff.

    :param ensemble: EnsembleData object containing the conformers.
    :param config: PartsConfig object with configuration settings.
    :param client: dask.distributed.Client for parallel execution.
    :param cut: Whether to apply cutting conditions.
    :return: JSON-serializable dictionary of refinement results.
    """
    printf(h2("REFINEMENT"))

    config = config.model_validate(config, context={"check": "refinement"})

    # Setup processor and target
    proc = Factory[QmProc].create(config.refinement.prog, "3_REFINEMENT")

    contributions_dict = {conf.name: Contributions() for conf in ensemble}
    if not config.general.gas_phase and not config.refinement.gsolv_included:
        # Calculate Gsolv using qm
        job_config = SPJobConfig(
            copy_mo=config.general.copy_mo,
            func=config.refinement.func,
            basis=config.refinement.basis,
            grid=GridLevel.VERY_HIGH,
            template=config.refinement.template,
            gas_phase=False,
            solvent=config.general.solvent,
            sm=config.refinement.sm,
            # multitemp=config.general.multitemp,
            # trange=config.general.trange,
            temperature=config.general.temperature,
            paths=config.paths,
        )
        results = execute(
            ensemble.conformers,
            proc.gsolv,
            job_config,
            config.refinement.prog,
            "refinement",
            client,
            ignore_failed=config.general.ignore_failed,
            balance=config.general.balance,
            copy_mo=config.general.copy_mo,
        )
        if config.general.ignore_failed:
            ensemble.remove_conformers(lambda conf: conf.name not in results)

        for conf in ensemble:
            contributions_dict[conf.name].gsolv = results[conf.name].gsolv
            contributions_dict[conf.name].energy = results[conf.name].energy_gas
    else:
        # Run sp calculation
        sp_job_config = SPJobConfig(
            copy_mo=config.general.copy_mo,
            func=config.refinement.func,
            basis=config.refinement.basis,
            grid=GridLevel.VERY_HIGH,
            template=config.refinement.template,
            gas_phase=config.general.gas_phase,
            solvent=config.general.solvent,
            sm=config.refinement.sm,
            paths=config.paths,
        )
        sp_results = execute(
            ensemble.conformers,
            proc.sp,
            sp_job_config,
            config.refinement.prog,
            "refinement",
            client,
            ignore_failed=config.general.ignore_failed,
            balance=config.general.balance,
            copy_mo=config.general.copy_mo,
        )
        if config.general.ignore_failed:
            ensemble.remove_conformers(lambda conf: conf.name not in sp_results)

        for conf in ensemble:
            contributions_dict[conf.name].energy = sp_results[conf.name].energy

    if config.general.evaluate_rrho:
        # Run mRRHO calculation
        proc_xtb = Factory[XtbProc].create(Prog.XTB, "3_REFINEMENT")
        rrho_job_config = RRHOJobConfig(
            gfnv=config.screening.gfnv,
            paths=config.paths,
            **config.general.model_dump(),  # This will just let the constructor pick the key/value pairs it needs
        )
        rrho_results = execute(
            ensemble.conformers,
            proc_xtb.xtb_rrho,
            rrho_job_config,
            "xtb",
            "refinement",
            client,
            ignore_failed=config.general.ignore_failed,
            balance=config.general.balance,
            copy_mo=config.general.copy_mo,
        )
        if config.general.ignore_failed:
            ensemble.remove_conformers(lambda conf: conf.name not in rrho_results)

        for conf in ensemble:
            contributions_dict[conf.name].grrho = rrho_results[conf.name].energy

    if cut:
        # Prepare Boltzmann populations
        boltzmann_populations = ensemble.get_populations(config.general.temperature)
        ensemble.conformers.sort(
            key=lambda conf: boltzmann_populations[conf.name], reverse=True
        )

        # Threshold in cumulative % pop
        threshold = config.refinement.threshold

        ensemble.remove_conformers(
            cond=lambda conf: sum(
                boltzmann_populations[c.name]
                for c in ensemble.conformers[: ensemble.conformers.index(conf)]
            )
            > threshold
        )

    # Print/write out results
    results_dict = jsonify(ensemble, config.refinement)
    _write_results(ensemble, config, results_dict)

    return results_dict


def _write_results(
    ensemble: EnsembleData, config: PartsConfig, results_dict: dict[str, Any]
) -> None:
    """ """
    printf(h1("REFINEMENT SINGLE-POINT (+ mRRHO) RESULTS"))

    # column headers
    headers = [
        "CONF#",
        # "G (xTB)",
        # "ΔG (xTB)",
        "E (DFT)",
        "ΔGsolv",
        "GmRRHO",
        "Gtot",
        "ΔGtot",
        "Boltzmann weight",
    ]

    # column units
    units = [
        "",
        # "[Eh]",
        # "[kcal/mol]",
        "[Eh]",
        "[kcal/mol]",
        "[Eh]",
        "[Eh]",
        "[kcal/mol]",
        f"% at {config.general.temperature} K",
    ]

    # variables for printmap
    # minimal xtb single-point energy
    # if all(
    #     "xtb_gsolv" in self.data["results"][conf.name]
    #     for conf in self._ensemble.conformers
    # ):
    #     xtbmin = min(
    #         self.data["results"][conf.name]["xtb_gsolv"]["energy_xtb_gas"]
    #         for conf in self._ensemble.conformers
    #     )

    gtotmin = min(conf.gtot for conf in ensemble)

    boltzmann_populations = ensemble.get_populations(config.general.temperature)

    printmap = {
        "CONF#": lambda conf: conf.name,
        # "G (xTB)": lambda conf: (
        #     f"{gxtb[conf.name]:.6f}" if gxtb is not None else "---"
        # ),
        # "ΔG (xTB)": lambda conf: (
        #     f"{(gxtb[conf.name] - gxtbmin) * AU2KCAL:.2f}"
        #     if gxtb is not None and gxtbmin is not None
        #     else "---"
        # ),
        "E (DFT)": lambda conf: f"{conf.energy:.6f}",
        "ΔGsolv": lambda conf: (f"{conf.gsolv * AU2KCAL:.6f}"),
        "GmRRHO": lambda conf: (f"{conf.grrho:.6f}"),
        "Gtot": lambda conf: f"{conf.gtot:.6f}",
        "ΔGtot": lambda conf: f"{(conf.gtot - gtotmin) * AU2KCAL:.2f}",
        "Boltzmann weight": lambda conf: f"{boltzmann_populations[conf.name] * 100:.2f}",
    }

    rows = [[printmap[header](conf) for header in headers] for conf in ensemble]

    for i in range(len(headers)):
        headers[i] += "\n" + units[i]

    table = tabulate(
        rows,
        headers=headers,
        colalign=["left"] + ["center" for _ in headers[1:]],
        disable_numparse=True,
        numalign="decimal",
    )
    print(table, flush=True)

    # list the averaged free enthalpy of the ensemble
    lines = []
    lines.append("\nBoltzmann averaged free energy/enthalpy of ensemble:")
    lines.append(f"{'temperature /K:':<15} {'avE(T) /a.u.':>14} {'avG(T) /a.u.':>14}")

    # calculate averaged free enthalpy
    avG = sum([boltzmann_populations[conf.name] * conf.gtot for conf in ensemble])

    # calculate averaged free energy (?)
    avE = sum([boltzmann_populations[conf.name] * conf.energy for conf in ensemble])

    # append the lines for the free energy/enthalpy
    lines.append(
        f"{config.general.temperature:^15} {avE:>14.7f}  {avG:>14.7f}     <<==part3=="
    )
    lines.append("".ljust(int(PLENGTH), "-"))

    # lines.append(f">>> END of {self.__class__.__name__} <<<".center(PLENGTH, " ") + "\n")

    # Print everything
    for line in lines:
        printf(line)

    # write everything to a file
    filepath = Path("3_REFINEMENT.out")
    filepath.write_text(table + "\n".join(lines))

    # Additionally, write results in json format
    Path("3_REFINEMENT.json").write_text(json.dumps(results_dict, indent=4))

    ensemble.dump_xyz(Path("3_REFINEMENT.xyz"))


# TODO: generalize this
def jsonify(
    ensemble: EnsembleData,
    config: RefinementConfig,
    fields: Callable[[MoleculeData], dict[str, Any]] | None = None,
):
    """
    Convert ensemble data to JSON format for refinement results.

    :param ensemble: EnsembleData object.
    :type ensemble: EnsembleData
    :param config: RefinementConfig object.
    :type config: RefinementConfig
    :param fields: Optional callable to customize fields.
    :type fields: Callable[[MoleculeData], dict[str, Any]] | None
    :return: JSON-serializable dictionary.
    :rtype: dict[str, Any]
    """
    per_conf: Callable[[MoleculeData], dict[str, dict[str, float]]] = fields or (
        lambda conf: {
            conf.name: {
                "energy": conf.energy,
                "gsolv": conf.gsolv,
                "grrho": conf.grrho,
                "gtot": conf.gtot,
            }
        }
    )

    dump = DataDump(part_name="refinement")

    for conf in ensemble:
        dump.data.update(per_conf(conf))

    dump.settings = config.model_dump()

    return dump.model_dump(mode="json")



from dataclasses import dataclass, field

from ..molecules import Atom


@dataclass
class Result:
    """Base class for results returned by calculations."""

    mo_path: str | tuple[str, str] = ""


@dataclass
class SPResult(Result):
    """Results class for single-point calculations."""

    energy: float = 0.0


@dataclass
class GsolvResult(Result):
    """Results class for Gsolv calculations."""

    gsolv: float = 0.0
    energy_gas: float = 0.0
    energy_solv: float = 0.0


@dataclass
class RRHOResult(Result):
    """Results class for mRRHO calculations."""

    energy: float = 0.0
    rmsd: float = 0.0
    gibbs: dict[float, float] = field(default_factory=dict)
    enthalpy: dict[float, float] = field(default_factory=dict)
    entropy: dict[float, float] = field(default_factory=dict)
    symmetry: str = "c1"
    symnum: int = 1
    linear: bool = False


@dataclass
class OptResult(Result):
    """Results class for geometry optimizations."""

    # 'ecyc' contains the energies for all cycles, 'cycles' stores the number of required cycles
    ecyc: list[float] = field(default_factory=list)
    # 'gncyc' contains the gradient norms for all cycles
    gncyc: list[float] = field(default_factory=list)
    # 'energy' contains the final energy of the optimization (converged or unconverged)
    energy: float = 0.0
    # 'geom' stores the optimized geometry in GeometryData.xyz format
    geom: list[Atom] = field(default_factory=list)
    grad_norm: float = 0.0
    cycles: int = 0
    converged: bool = False


@dataclass
class NMRResult(Result):
    """Results class for NMR calculations."""

    # atom index mapped onto shielding value, starting from 0
    shieldings: list[tuple[int, float]] = field(default_factory=list)
    # two atom indices mapped onto coupling value, starting from 0
    couplings: list[tuple[tuple[int, int], float]] = field(default_factory=list)


@dataclass
class RotResult(Result):
    """Results class for optical rotation calculations."""

    rotations_velocity: list[tuple[float, float]] = field(default_factory=list)
    rotations_length: list[tuple[float, float]] = field(default_factory=list)


@dataclass
class UVVisResult(Result):
    """Results class for UVVis calculations."""

    excitations: list[dict[str, float]] = field(default_factory=list)


@dataclass
class MetaData:
    """Metadata for job execution results."""

    conf_name: str
    success: bool = False
    error: str = ""



"""
Generic processor class. Only abstract methods: sp, opt (single-point and geometry optimization).
Also implements important methods inherited by all other processors, e.g. making external program calls and creating job dirs.
"""

from abc import abstractmethod
import os
from pathlib import Path
import subprocess
import time
from typing import final
from dask.distributed import Variable, Lock


from ..config.job_config import (
    SPJobConfig,
    OptJobConfig,
)
from .job import (
    JobContext,
)
from .results import (
    OptResult,
    MetaData,
    SPResult,
)
from ..params import (
    WARNLEN,
    Prog,
    ENVIRON,
)
from ..utilities import printf
from ..logging import setup_logger

logger = setup_logger(__name__)


class GenericProc:
    """
    Generic processor class
    """

    progname: Prog = Prog.GENERIC

    # rotational entropy from symmetry
    # https://cccbdb.nist.gov/thermo.asp
    _rot_sym_num = {
        "c1": 1,
        "ci": 1,
        "cs": 1,
        "c2": 2,
        "c3": 3,
        "c4": 4,
        "c5": 5,
        "c6": 6,
        "c7": 7,
        "c8": 8,
        "c9": 9,
        "c10": 10,
        "c11": 11,
        "s4": 2,
        "s6": 3,
        "s8": 4,
        "d2": 4,
        "d3": 6,
        "d4": 8,
        "d5": 10,
        "d6": 12,
        "d7": 14,
        "d8": 16,
        "d9": 18,
        "d10": 20,
        "t": 12,
        "th": 12,
        "td": 12,
        "o": 24,
        "oh": 24,
        "ih": 60,
    }

    # @staticmethod
    # @final
    # def _run(f):
    #     """
    #     Wrapper function to manage resources and create job directory.
    #
    #     :param f: Callable that returns a tuple (results and metadata).
    #     :return: The wrapped function.
    #     """
    #
    #     @functools.wraps(f)
    #     def wrapper(
    #         self,
    #         job: JobContext,
    #         job_config,
    #         **kwargs,
    #     ):
    #         jobtype = f.__name__
    #         jobdir = Path(self._workdir) / job.conf.name / jobtype
    #         jobdir.mkdir(exist_ok=True, parents=True)
    #
    #         logger.info(
    #             f"{f'worker{os.getpid()}:':{WARNLEN}}Running "
    #             + f"{jobtype} calculation using {self.__class__.__name__} in {jobdir} on {job.omp} cores using {self.progname.value}."
    #         )
    #         printf(
    #             f"Running {jobtype} calculation for {job.conf.name} using {self.progname.value}."
    #         )
    #
    #         result, meta = f(self, job, jobdir, job_config, **kwargs)
    #
    #         return result, meta
    #
    #     return wrapper
    #
    @final
    def _setup(self, job: JobContext, jobtype: str):
        """
        Setup function to create job directory and print info before task execution.

        :param job: Job context
        :type job: JobContext
        :param jobtype: Type of job
        :type jobtype: str
        :returns: Path to the job directory
        :rtype: Path
        """
        jobdir = Path(self._workdir) / job.conf.name / jobtype
        jobdir.mkdir(exist_ok=True, parents=True)

        logger.info(
            f"{f'worker{os.getpid()}:':{WARNLEN}}Running "
            + f"{jobtype} calculation using {self.__class__.__name__} in {jobdir} on {job.omp} cores using {self.progname.value}."
        )
        printf(
            f"Running {jobtype} calculation for {job.conf.name} using {self.progname.value}."
        )

        return jobdir

    def __init__(self, workdir: Path):
        """
        QM processor base class containing only xtb-related functions.

        :param workdir: Working directory.
        """
        self._workdir: Path = workdir
        self._lock: Lock = Lock()
        self.cancel_var: Variable | None = None

    @final
    def _make_call(
        self,
        call: list[str],
        outputpath: str,
        jobdir: str | Path,
        env: dict[str, str] | None = None,
    ) -> tuple[int, str]:
        """
        Make a call to an external program and write output into outputfile.

        :param call: List containing the call args to the external program.
        :param outputpath: Path to the outputfile.
        :param jobdir: Path to the jobdir.
        :return: Tuple of (returncode of the external program or -1 in case of exception, stderr output).
        """
        try:
            # call external program and write output into outputfile
            with open(outputpath, "w", newline=None) as outputfile:
                logger.debug(f"{f'worker{os.getpid()}:':{WARNLEN}}Running {call}...")

                # create subprocess for external program
                sub = subprocess.Popen(
                    call,
                    shell=False,
                    stderr=subprocess.PIPE,
                    cwd=jobdir,
                    stdout=outputfile,
                    env=env or ENVIRON,
                )

                logger.debug(
                    f"{f'worker{os.getpid()}:':{WARNLEN}}Started (PID: {sub.pid})."
                )

                # Cooperative cancellation using dask Variable
                while sub.poll() is None:
                    with self._lock:
                        if self.cancel_var and self.cancel_var.get():
                            logger.info(
                                f"{f'worker{os.getpid()}:':{WARNLEN}}Cancelling subprocess {sub.pid} due to cancellation signal."
                            )
                            sub.terminate()
                            break
                    time.sleep(0.1)

                # wait for process to finish
                _, stderr = sub.communicate()
                errors = stderr.decode(errors="replace")
                returncode = sub.returncode
                if returncode != 0:
                    logger.info(
                        f"{f'worker{os.getpid()}:':{WARNLEN}} Returncode: {returncode} Errors:\n{errors}"
                    )

            logger.debug(f"{f'worker{os.getpid()}:':{WARNLEN}}Done.")

            return returncode, errors
        except Exception as e:
            logger.error(f"{f'worker{os.getpid()}:':{WARNLEN}}Error: {e}")
            return -1, str(e)

    @final
    def _get_sym_num(self, sym: str | None = None, linear: bool = False) -> int:
        """
        Get rotational symmetry number from Schoenfließ symbol.

        :param sym: Symmetry symbol.
        :param linear: Whether the molecule is linear.
        :return: Symmetry number.
        """
        if sym is None:
            sym = "c1"
        symnum = 1
        if linear and "c" in sym.lower()[0]:
            symnum = 1
            return symnum
        elif linear and "d" in sym.lower()[0]:
            symnum = 2
            return symnum
        for key in self._rot_sym_num:
            if key in sym.lower():
                symnum = self._rot_sym_num.get(key, 1)
                break
        return symnum

    @abstractmethod
    def sp(self, job: JobContext, config: SPJobConfig) -> tuple[SPResult, MetaData]:
        """
        Perform single-point calculation.

        :param job: Parallel job.
        :param config: SP configuration.
        :return: Tuple of (SP result, metadata).
        """
        raise NotImplementedError

    @abstractmethod
    def opt(self, job: JobContext, config: OptJobConfig) -> tuple[OptResult, MetaData]:
        """
        Perform geometry optimization.

        :param job: Parallel job.
        :param config: Optimization configuration.
        :return: Tuple of (optimization result, metadata).
        """
        raise NotImplementedError
